from typing import Dict, Optional, Union
import pathlib

import numpy as np
import pandas as pd
import xtgeo

try:
    import _roxar  # type: ignore
except ModuleNotFoundError:

    class MockRoxar:
        @property
        def Project(self):
            pass

    _roxar = MockRoxar()


def extract_grid_zone_tops(
    project: Optional[_roxar.Project] = None,
    well_list: Optional[list] = None,
    logrun: str = "log",
    trajectory: str = "Drilled trajectory",
    gridzonelog: str = None,
    mdlogname: str = None,
    grid: str = None,
    zone_param: str = None,
    alias_file: str = None,
    rms_name: str = "RMS_WELL_NAME",
    ecl_name: str = "ECLIPSE_WELL_NAME",
) -> pd.DataFrame:
    """
    Function for extracting top and base from gridzones, both in TVD and MD.
    A pandas dataframe will be returned.

    Users can either input a pre-generated gridzonelog or a grid and a zone parameter
    for computing the gridzonelog.

    The function works both inside RMS and outside with file input. If input from files,
    and a MD log is not present in the well a quasi md log will be computed and used.
    """
    use_gridzonelog = False if gridzonelog is None else True

    if not use_gridzonelog:
        if grid is not None and zone_param is not None:
            if project is not None:
                mygrid = xtgeo.grid_from_roxar(project, grid)
                gridzones = xtgeo.gridproperty_from_roxar(project, grid, zone_param)
            else:
                mygrid = xtgeo.grid_from_file(grid)
                gridzones = xtgeo.gridproperty_from_file(zone_param, grid=mygrid)
            gridzones.name = "Zone"
        else:
            raise ValueError("Specify either 'gridzonelog' or 'grid' and 'zone_param")

    dfs = []

    if well_list is None:
        well_list = []

    for well in well_list:
        try:
            if project is not None:
                xtg_well = xtgeo.well_from_roxar(
                    project,
                    str(well),
                    trajectory=trajectory,
                    logrun=logrun,
                    inclmd=True,
                )
            else:
                xtg_well = xtgeo.well_from_file(str(well), mdlogname=mdlogname)
                # quasi md log will be computed
                xtg_well.geometrics()
        except (ValueError, KeyError):
            continue

        # if no gridzonelog create one from the zone parameter
        if not use_gridzonelog:
            xtg_well.get_gridproperties(gridzones, mygrid)
            gridzonelog = "Zone_model"

        if xtg_well.dataframe[gridzonelog].isnull().values.all():
            continue

        # Set gridzonelog as zonelog and extract zonation tops from it
        xtg_well.zonelogname = gridzonelog
        dframe = xtg_well.get_zonation_points(top_prefix="", use_undef=True)

        dframe.rename(
            columns={
                "Z_TVDSS": "TOP_TVD",
                xtg_well.mdlogname: "TOP_MD",
                "Zone": "ZONE_CODE",
                "WellName": "WELL",
            },
            inplace=True,
        )
        # find deepest point in well while in grid
        df_max = (
            xtg_well.dataframe[["Z_TVDSS", xtg_well.mdlogname, gridzonelog]]
            .dropna()
            .sort_values(by=xtg_well.mdlogname)
        )
        # create base picks also
        dframe["BASE_TVD"] = dframe["TOP_TVD"].shift(-1)
        dframe["BASE_MD"] = dframe["TOP_MD"].shift(-1)
        dframe.at[dframe.index[-1], "BASE_TVD"] = df_max.iloc[-1]["Z_TVDSS"]
        dframe.at[dframe.index[-1], "BASE_MD"] = df_max.iloc[-1][xtg_well.mdlogname]
        # adjust zone values to get correct zone information
        dframe["ZONE_CODE"] = shift_zone_values(dframe["ZONE_CODE"].values.copy())
        dframe["ZONE"] = (
            dframe["ZONE_CODE"]
            .map(xtg_well.get_logrecord(xtg_well.zonelogname))
            .fillna("Outside")
        )
        dfs.append(dframe.drop(columns=["TopName", "Q_INCL", "Q_AZI"]))

    df = pd.concat(dfs)
    if alias_file is not None:
        well_dict = make_alias_dict(alias_file, rms_name, ecl_name)
        df["WELL"] = df["WELL"].replace(well_dict)
    return df


def shift_zone_values(zvals: np.ndarray) -> np.ndarray:
    for idx, _zval in enumerate(zvals):
        if idx == len(zvals) - 1:
            continue
        if zvals[idx] == zvals[idx + 1]:
            zvals[idx + 1] = zvals[idx + 1] - 1
    return zvals


def make_alias_dict(
    alias_file: Union[str, pathlib.Path],
    rms_name: str = "RMS_WELL_NAME",
    ecl_name: str = "ECLIPSE_WELL_NAME",
) -> Dict[str, str]:
    """
    Create a correspondance dictionary so that well_dict[ <RMS wellname> ]
    = <Eclipse wellname>
    """
    df = pd.read_csv(alias_file, index_col=rms_name)
    well_dict = df.to_dict()
    well_dict = well_dict[ecl_name]

    return well_dict



from ._blockedwells_vs_gridprops import BlockedWellsVsGridProperties
from ._grid_quality import GridQuality
from ._grid_statistics import GridStatistics
from ._wellzonation_vs_grid import WellZonationVsGrid
from .qcforward import (
    blockedwells_vs_gridproperties,
    grid_quality,
    grid_statistics,
    wellzonation_vs_grid,
)

__all__ = [
    "wellzonation_vs_grid",
    "WellZonationVsGrid",
    "grid_statistics",
    "GridStatistics",
    "grid_quality",
    "GridQuality",
    "BlockedWellsVsGridProperties",
    "blockedwells_vs_gridproperties",
]



"""Module for reading excel file with input for generation of
a design matrix and converting to an OrderedDict that can be
read by fmu.tools.DesignMatrix.generate
"""
from collections import OrderedDict
import warnings
import numpy as np
import openpyxl
import pandas as pd
import yaml

SEEDS_DEPRECATION_WARNING = """
The keyword "seeds" in the "general_input" sheet is changed
to "rms_seeds", and will be deprecated in a future version.
Please update your excel spreadsheet
"""


def excel2dict_design(input_filename, sheetnames=None):
    """Read excel file with input to design setup
    Currently only specification of
    onebyone design is implemented

    Args:
        input_filename (str): Name of excel input file
        sheetnames (dict): Dictionary of worksheet names to load
            information from. Supported keys: general_input, defaultvalues,
            and designinput.

    Returns:
        OrderedDict on format for DesignMatrix.generate
    """
    if not sheetnames or "general_input" not in sheetnames:
        gen_input_sheet = _find_geninput_sheetname(input_filename)
    else:
        gen_input_sheet = sheetnames["general_input"]

    generalinput = pd.read_excel(
        input_filename, gen_input_sheet, header=None, index_col=0, engine="openpyxl"
    )
    generalinput.dropna(axis=0, how="all", inplace=True)
    generalinput.dropna(axis=1, how="all", inplace=True)

    if str(generalinput[1]["designtype"]) == "onebyone":
        returndict = _excel2dict_onebyone(input_filename, sheetnames)
    else:
        raise ValueError(
            "Generation of DesignMatrix only "
            "implemented for type onebyone "
            "In general_input designtype was "
            "set to {}".format(str(generalinput[1]["designtype"]))
        )
    return returndict


def inputdict_to_yaml(inputdict, filename):
    """Write inputdict to yaml format

    Args:
        inputdict (OrderedDict)
        filename (str): path for where to write file
    """
    stream = open(filename, "w")
    yaml.dump(inputdict, stream)


def _find_geninput_sheetname(input_filename):
    """Finding general input sheet, allowing for name
    variations."""
    xlsx = openpyxl.load_workbook(input_filename, read_only=True, keep_links=False)
    sheets = xlsx.sheetnames
    general_input_sheet = []
    for sheet in sheets:
        if sheet in [
            "general_input",
            "generalinput",
            "GeneralInput",
            "Generalinput",
            "General_Input",
            "General_input",
        ]:
            general_input_sheet.append(sheet)

    if len(general_input_sheet) > 1:
        raise ValueError(
            "More than one sheet with general input"
            "Sheetnames are {} ".format(general_input_sheet)
        )

    if not general_input_sheet:
        raise ValueError(
            "No general_input sheet provided in Excel file {} "
            "".format(input_filename)
        )

    return general_input_sheet[0]


def _find_onebyone_defaults_sheet(input_filename):
    """Finds correct sheet name for default values to use when parsing
    excel file.

    Returns:
        string, name of a sheet in the excel file
    """
    xlsx = openpyxl.load_workbook(input_filename, read_only=True, keep_links=False)
    sheets = xlsx.sheetnames

    default_values_sheet = []

    for sheet in sheets:
        if sheet in [
            "default_values",
            "defaultvalues",
            "DefaultValues",
            "Defaultvalues",
            "Default_Values",
            "Default_values",
        ]:
            default_values_sheet.append(sheet)
    if len(default_values_sheet) > 1:
        raise ValueError(
            "More than one sheet with default values"
            "Sheetnames are {} ".format(default_values_sheet)
        )

    if len(default_values_sheet) == []:
        raise ValueError(
            "No defaultvalues sheet provided in Excel file {} "
            "".format(input_filename)
        )

    return default_values_sheet[0]


def _find_onebyone_input_sheet(input_filename):
    """Finds correct sheet name for input to use when parsing excel file.

    Returns:
        string, name of a sheet in the excel file
    """
    xlsx = openpyxl.load_workbook(input_filename, read_only=True, keep_links=False)
    sheets = xlsx.sheetnames

    design_input_sheet = []

    for sheet in sheets:
        if sheet in [
            "design_input",
            "designinput",
            "DesignInput",
            "Designinput",
            "Design_Input",
            "Design_input",
        ]:
            design_input_sheet.append(sheet)
    if len(design_input_sheet) > 1:
        raise ValueError(
            "More than one sheet with design input"
            "Sheetnames are {} ".format(design_input_sheet)
        )

    if not design_input_sheet:
        raise ValueError(
            "No designinput sheet provided in Excel file {} " "".format(input_filename)
        )
    return design_input_sheet[0]


def _check_designinput(dsgn_input):
    """Checks for valid input in designinput sheet"""

    # Check for duplicate sensnames
    sensitivity_names = []
    for row in dsgn_input.itertuples():
        if _has_value(row.sensname):
            if row.sensname in sensitivity_names:
                raise ValueError(
                    "sensname '{}' was found on more than one row in designinput "
                    "sheet. Two sensitivities can not share the same sensname. "
                    "Please correct this and rerun".format(row.sensname)
                )
            sensitivity_names.append(row.sensname)


def _check_for_mixed_sensitivities(sens_name, sens_group):
    """Checks for valid input in designinput sheet. A sensitivity cannot contain
    two different sensitivity types"""

    types = sens_group.groupby("type", sort=False)
    if len(types) > 1:
        raise ValueError(
            "The sensitivity with sensname '{}' in designinput sheet contains more "
            "than one sensitivity type. For each sensname all parameters must be "
            "specified using the same type (seed, scenario, dist, ref, background, "
            "extern)".format(sens_name)
        )


def _excel2dict_onebyone(input_filename, sheetnames=None):
    """Reads spesification for onebyone design

    Args:
        input_filename(path): path to excel workbook
        sheetnames (dict): Dictionary of worksheet names to load
            information from. Supported keys: general_input, defaultvalues,
            and designinput.

    Returns:
        OrderedDict on format for DesignMatrix.generate
    """

    seedname = "RMS_SEED"
    inputdict = OrderedDict()

    if not sheetnames or "general_input" not in sheetnames:
        gen_input_sheet = _find_geninput_sheetname(input_filename)
    else:
        gen_input_sheet = sheetnames["general_input"]

    if not sheetnames or "designinput" not in sheetnames:
        design_inp_sheet = _find_onebyone_input_sheet(input_filename)
    else:
        design_inp_sheet = sheetnames["designinput"]

    if not sheetnames or "defaultvalues" not in sheetnames:
        default_val_sheet = _find_onebyone_defaults_sheet(input_filename)
    else:
        default_val_sheet = sheetnames["defaultvalues"]

    # Read general input
    generalinput = pd.read_excel(
        input_filename, gen_input_sheet, header=None, index_col=0, engine="openpyxl"
    )
    generalinput.dropna(axis=0, how="all", inplace=True)
    generalinput.dropna(axis=1, how="all", inplace=True)

    inputdict["designtype"] = generalinput[1]["designtype"]

    if "rms_seeds" in generalinput[1]:
        if str(generalinput[1]["rms_seeds"]) == "None":
            inputdict["seeds"] = None
        else:
            inputdict["seeds"] = generalinput[1]["rms_seeds"]
    elif "seeds" in generalinput[1]:
        warnings.simplefilter("always")
        warnings.warn(SEEDS_DEPRECATION_WARNING, DeprecationWarning)
        if str(generalinput[1]["seeds"]) == "None":
            inputdict["seeds"] = None
        else:
            inputdict["seeds"] = generalinput[1]["seeds"]
    else:
        inputdict["seeds"] = None

    if "repeats" not in generalinput[1]:
        raise LookupError('"repeats" must be specified in general_input sheet')

    inputdict["repeats"] = generalinput[1]["repeats"]

    if "distribution_seed" in generalinput[1]:
        if str(generalinput[1]["distribution_seed"]) == "None":
            inputdict["distribution_seed"] = None
        else:
            inputdict["distribution_seed"] = generalinput[1]["distribution_seed"]
    else:
        inputdict["distribution_seed"] = None

    # Read background
    if "background" in generalinput.index:
        inputdict["background"] = OrderedDict()
        if generalinput[1]["background"].endswith("csv") or generalinput[1][
            "background"
        ].endswith("xlsx"):
            inputdict["background"]["extern"] = generalinput[1]["background"]
        elif str(generalinput[1]["background"]) == "None":
            inputdict["background"] = None
        else:
            inputdict["background"] = _read_background(
                input_filename, generalinput[1]["background"]
            )
    else:
        inputdict["background"] = None

    # Read default values
    inputdict["defaultvalues"] = _read_defaultvalues(input_filename, default_val_sheet)

    # Read input for sensitivities
    inputdict["sensitivities"] = OrderedDict()
    designinput = pd.read_excel(input_filename, design_inp_sheet, engine="openpyxl")
    designinput.dropna(axis=0, how="all", inplace=True)
    designinput = designinput.loc[:, ~designinput.columns.str.contains("^Unnamed")]

    # First column with parameter names should have spaces stripped,
    # but we need to preserve NaNs:
    not_nan_sensnames = ~designinput["sensname"].isnull()
    designinput.loc[not_nan_sensnames, "sensname"] = (
        designinput.loc[not_nan_sensnames, "sensname"].astype(str).str.strip()
    )

    _check_designinput(designinput)

    designinput["sensname"].fillna(method="ffill", inplace=True)

    # Read dependencies
    if "dependencies" in designinput.keys():
        inputdict["dependencies"] = OrderedDict()
        for row in designinput.itertuples():
            if _has_value(row.dependencies):
                inputdict["dependencies"][row.param_name] = _read_dependencies(
                    input_filename, row.dependencies, row.param_name
                )

    # Read decimals
    if "decimals" in designinput.keys():
        inputdict["decimals"] = OrderedDict()
        for row in designinput.itertuples():
            if _has_value(row.decimals) and _is_int(row.decimals):
                inputdict["decimals"][row.param_name] = int(row.decimals)

    grouped = designinput.groupby("sensname", sort=False)

    # Read each sensitivity
    for sensname, group in grouped:

        _check_for_mixed_sensitivities(sensname, group)

        sensdict = OrderedDict()

        if group["type"].iloc[0] == "ref":
            sensdict["senstype"] = "ref"

        elif group["type"].iloc[0] == "background":
            sensdict["senstype"] = "background"

        elif group["type"].iloc[0] == "seed":
            sensdict["seedname"] = seedname
            sensdict["senstype"] = "seed"
            if _has_value(group["param_name"].iloc[0]):
                sensdict["parameters"] = _read_constants(group)
            else:
                sensdict["parameters"] = None

        elif group["type"].iloc[0] == "scenario":
            sensdict = _read_scenario_sensitivity(group)
            sensdict["senstype"] = "scenario"

        elif group["type"].iloc[0] == "dist":
            sensdict["senstype"] = "dist"
            sensdict["parameters"] = _read_dist_sensitivity(group)
            sensdict["correlations"] = None
            if "corr_sheet" in group.keys():
                sensdict["correlations"] = _read_correlations(group, input_filename)

        elif group["type"].iloc[0] == "extern":
            sensdict["extern_file"] = str(group["extern_file"].iloc[0])
            sensdict["senstype"] = "extern"
            sensdict["parameters"] = list(group["param_name"])

        else:
            raise ValueError(
                "Sensitivity {} does not have a valid sensitivity type".format(sensname)
            )

        if "numreal" in group.keys():
            if _has_value(group["numreal"].iloc[0]):
                # Using default number of realisations:
                # 'repeats' from general_input sheet
                sensdict["numreal"] = int(group["numreal"].iloc[0])

        inputdict["sensitivities"][str(sensname)] = sensdict

    return inputdict


def _read_defaultvalues(filename, sheetname):
    """Reads defaultvalues, also used as values for
    reference/base case

    Args:
        filename(path): path to excel file
        sheetname (string): name of defaultsheet

    Returns:
        OrderedDict with defaultvalues (parameter, value)
    """
    default_dict = OrderedDict()
    default_df = pd.read_excel(
        filename, sheetname, header=0, index_col=0, engine="openpyxl"
    )
    default_df.dropna(axis=0, how="all", inplace=True)
    default_df = default_df.loc[:, ~default_df.columns.str.contains("^Unnamed")]

    # Strip spaces before and after parameter names, if they are there
    # it is probably invisible user errors in Excel.
    default_df.index = [
        paramname.strip() if isinstance(paramname, str) else paramname
        for paramname in default_df.index
    ]
    for row in default_df.itertuples():
        if str(row[0]) in default_dict.keys():
            print(
                "WARNING: The default value '{}' "
                "is listed twice in the sheet '{}'. "
                "Only the first entry will be used in output file".format(
                    row[0], sheetname
                )
            )
        else:
            default_dict[str(row[0])] = row[1]
    return default_dict


def _read_dependencies(filename, sheetname, from_parameter):
    """Reads parameters that are set from other parameters

    Args:
        filename(path): path to excel file
        sheetname (string): name of dependency sheet

    Returns:
        OrderedDict with design parameter, dependent parameters
        and values
    """
    depend_dict = OrderedDict()
    depend_df = pd.read_excel(
        filename, sheetname, dtype=str, na_values="", engine="openpyxl"
    )
    depend_df.dropna(axis=0, how="all", inplace=True)
    depend_df = depend_df.loc[:, ~depend_df.columns.str.contains("^Unnamed")]

    if from_parameter in depend_df.keys():
        depend_dict["from_values"] = depend_df[from_parameter].tolist()
        depend_dict["to_params"] = OrderedDict()
        for key in depend_df.keys():
            if key != from_parameter:
                depend_dict["to_params"][key] = depend_df[key].tolist()
    else:
        raise ValueError(
            "Parameter {} specified to have derived parameters, "
            "but the sheet specifying the dependencies {} does "
            "not contain the input parameter. ".format(from_parameter, sheetname)
        )
    return depend_dict


def _read_background(inp_filename, bck_sheet):
    """Reads excel sheet with background parameters and distributions

    Args:
        inp_filename (path): path to Excel workbook
        bck_sheet (str): name of sheet with background parameters

    Returns:
        OrderedDict with parameter names and distributions
    """
    backdict = OrderedDict()
    paramdict = OrderedDict()
    bck_input = pd.read_excel(inp_filename, bck_sheet, engine="openpyxl")
    bck_input.dropna(axis=0, how="all", inplace=True)
    bck_input = bck_input.loc[:, ~bck_input.columns.str.contains("^Unnamed")]

    backdict["correlations"] = None
    if "corr_sheet" in bck_input.keys():
        backdict["correlations"] = _read_correlations(bck_input, inp_filename)

    if "dist_param1" not in bck_input.columns.values:
        bck_input["dist_param1"] = float("NaN")
    if "dist_param2" not in bck_input.columns.values:
        bck_input["dist_param2"] = float("NaN")
    if "dist_param3" not in bck_input.columns.values:
        bck_input["dist_param3"] = float("NaN")
    if "dist_param4" not in bck_input.columns.values:
        bck_input["dist_param4"] = float("NaN")

    for row in bck_input.itertuples():
        if not _has_value(row.param_name):
            raise ValueError(
                "Background parameters specified "
                "where one line has empty parameter "
                "name "
            )
        if not _has_value(row.dist_param1):
            raise ValueError(
                "Parameter {} has been input "
                "in background sheet but with empty "
                "first distribution parameter ".format(row.param_name)
            )
        if not _has_value(row.dist_param2) and _has_value(row.dist_param3):
            raise ValueError(
                "Parameter {} has been input in "
                "background sheet with "
                'value for "dist_param3" while '
                '"dist_param2" is empty. This is not '
                "allowed".format(row.param_name)
            )
        if not _has_value(row.dist_param3) and _has_value(row.dist_param4):
            raise ValueError(
                "Parameter {} has been input in "
                "background sheet with "
                'value for "dist_param4" while '
                '"dist_param3" is empty. This is not '
                "allowed".format(row.param_name)
            )
        distparams = [
            item
            for item in [
                row.dist_param1,
                row.dist_param2,
                row.dist_param3,
                row.dist_param4,
            ]
            if _has_value(item)
        ]
        if "corr_sheet" in bck_input.keys():
            if not _has_value(row.corr_sheet):
                corrsheet = None
            else:
                corrsheet = row.corr_sheet
        else:
            corrsheet = None
        paramdict[str(row.param_name)] = [str(row.dist_name), distparams, corrsheet]
    backdict["parameters"] = paramdict

    if "decimals" in bck_input.keys():
        decimals = OrderedDict()
        for row in bck_input.itertuples():
            if _has_value(row.decimals) and _is_int(row.decimals):
                decimals[row.param_name] = int(row.decimals)
        backdict["decimals"] = decimals

    return backdict


def _read_scenario_sensitivity(sensgroup):
    """Reads parameters and values
    for scenario sensitivities
    """
    sdict = OrderedDict()
    sdict["cases"] = OrderedDict()
    casedict1 = OrderedDict()
    casedict2 = OrderedDict()

    if not _has_value(sensgroup["senscase1"].iloc[0]):
        raise ValueError(
            "Sensitivity {} has been input "
            "as a scenario sensitivity, but "
            "without a name in senscase1 column.".format(sensgroup["sensname"].iloc[0])
        )

    for row in sensgroup.itertuples():
        if not _has_value(row.param_name):
            raise ValueError(
                "Scenario sensitivity {} specified "
                "where one line has empty parameter "
                "name ".format(row.sensname)
            )
        if not _has_value(row.value1):
            raise ValueError(
                "Parameter {} har been input "
                'as type "scenario" but with empty '
                "value in value1 column ".format(row.param_name)
            )
        casedict1[str(row.param_name)] = row.value1

    if _has_value(sensgroup["senscase2"].iloc[0]):
        for row in sensgroup.itertuples():
            if not _has_value(row.value2):
                raise ValueError(
                    "Sensitivity {} has been input "
                    "with a name in senscase2 column "
                    "but without a value for parameter {} "
                    "in value2 column.".format(
                        sensgroup["sensname"].iloc[0], row.param_name
                    )
                )
            casedict2[str(row.param_name)] = row.value2
        sdict["cases"][str(sensgroup["senscase1"].iloc[0])] = casedict1
        sdict["cases"][str(sensgroup["senscase2"].iloc[0])] = casedict2
    else:
        for row in sensgroup.itertuples():
            if _has_value(row.value2):
                raise ValueError(
                    "Sensitivity {} has been input "
                    "with a value for parameter {} "
                    "in value2 column "
                    "but without a name for the scenario "
                    "in senscase2 column.".format(
                        sensgroup["sensname"].iloc[0], row.param_name
                    )
                )
        sdict["cases"][str(sensgroup["senscase1"].iloc[0])] = casedict1
    return sdict


def _read_constants(sensgroup):
    """Reads constants to be used together with
    seed sensitivity"""
    if "dist_param1" not in sensgroup.columns.values:
        sensgroup["dist_param1"] = float("NaN")
    paramdict = OrderedDict()
    for row in sensgroup.itertuples():
        if not _has_value(row.dist_param1):
            raise ValueError(
                "Parameter name {} has been input "
                'in a sensitivity of type "seed". \n'
                "If {} was meant to be the name of "
                "the seed parameter, this is "
                "unfortunately not allowed. "
                "The seed parameter name is standardised "
                "to RMS_SEED and should not be specified.\n "
                "If you instead meant to specify a constant "
                "value for another parameter in the seed "
                'sensitivity, please remember "const" in '
                'dist_name and a value in "dist_param1". '.format(
                    row.param_name, row.param_name
                )
            )
        distparams = row.dist_param1
        paramdict[str(row.param_name)] = [str(row.dist_name), distparams]
    return paramdict


def _read_dist_sensitivity(sensgroup):
    """Reads parameters and distributions
    for monte carlo sensitivities
    """
    if "dist_param1" not in sensgroup.columns.values:
        sensgroup["dist_param1"] = float("NaN")
    if "dist_param2" not in sensgroup.columns.values:
        sensgroup["dist_param2"] = float("NaN")
    if "dist_param3" not in sensgroup.columns.values:
        sensgroup["dist_param3"] = float("NaN")
    if "dist_param4" not in sensgroup.columns.values:
        sensgroup["dist_param4"] = float("NaN")
    paramdict = OrderedDict()
    for row in sensgroup.itertuples():
        if not _has_value(row.param_name):
            raise ValueError(
                "Dist sensitivity {} specified "
                "where one line has empty parameter "
                "name ".format(row.sensname)
            )
        if not _has_value(row.dist_param1):
            raise ValueError(
                "Parameter {} has been input "
                'as type "dist" but with empty '
                "first distribution parameter ".format(row.param_name)
            )
        if not _has_value(row.dist_param2) and _has_value(row.dist_param3):
            raise ValueError(
                "Parameter {} has been input with "
                'value for "dist_param3" while '
                '"dist_param2" is empty. This is not '
                "allowed".format(row.param_name)
            )
        if not _has_value(row.dist_param3) and _has_value(row.dist_param4):
            raise ValueError(
                "Parameter {} has been input with "
                'value for "dist_param4" while '
                '"dist_param3" is empty. This is not '
                "allowed".format(row.param_name)
            )
        distparams = [
            item
            for item in [
                row.dist_param1,
                row.dist_param2,
                row.dist_param3,
                row.dist_param4,
            ]
            if _has_value(item)
        ]
        if "corr_sheet" in sensgroup.keys():
            if not _has_value(row.corr_sheet):
                corrsheet = None
            else:
                corrsheet = row.corr_sheet
        else:
            corrsheet = None
        paramdict[str(row.param_name)] = [str(row.dist_name), distparams, corrsheet]

    return paramdict


def _read_correlations(sensgroup, inputfile):

    if "corr_sheet" in sensgroup.keys():
        if not sensgroup["corr_sheet"].dropna().empty:
            correlations = OrderedDict()
            correlations["inputfile"] = inputfile
            correlations["sheetnames"] = []
            for _index, row in sensgroup.iterrows():
                if _has_value(row["corr_sheet"]):
                    if row["corr_sheet"] not in correlations["sheetnames"]:
                        correlations["sheetnames"].append(row["corr_sheet"])
        else:
            correlations = None
    else:
        correlations = None

    return correlations


def _has_value(value):
    """Returns False only if the argument is np.nan"""
    try:
        return not np.isnan(value)
    except TypeError:
        return True


def _is_int(teststring):
    """Test if a string can be parsed as a float"""
    try:
        if not np.isnan(int(teststring)):
            if (float(teststring) % 1) == 0:
                return True
            return False
        return False  # It was a "number", but it was NaN.
    except ValueError:
        return False



"""
create_rft_ertobs creates txt, obs and welldatefile for usage together
with GENDATA in ERT assisted history match runs.

This script should be run from within RMS to be able to interpolate
along well trajectories::

    from fmu.tools.rms import create_rft_ertobs

    SETUP = {...yourconfiguration...}  # A Python dictionary

    create_rft_ertobs.main(SETUP)

Required keys in the SETUP dictionary:

* input_file or input_dframe: Path to CSV file with data for wellnames, dates
  wellpoint coordinates and pressure values.

Optional keys:

* rft_prefix: Eclipse well names will be prefixed with this string
* gridname: Name of grid in RMS project (for verification)
* zonename: Name of zone parameter in RMS grid (if zone names should be
  verified)
* welldatefile: Name of welldatefile, written to exportdir.
  Defaults to "well_date_rft.txt"

Result:

* ``txt`` files written to ert/input/observations
* ``obs`` files for each well written to ert/input/observations
* ``well_date_rft.txt``
"""
import logging
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import numpy as np
import pandas as pd
from scipy.interpolate import CubicSpline

logger = logging.getLogger(__name__)

SUPPORTED_OPTIONS = [
    "absolute_error",
    "alias",
    "alias_file",
    "ecl_name",
    "exportdir",
    "gridname",
    "input_dframe",
    "input_file",
    "interpolation",
    "loglevel",
    "project",
    "relative_error",
    "rft_prefix",
    "rms_name",
    "trajectory_name",
    "verbose",
    "welldatefile",  # WELL_AND_TIME_FILE in semeio
    "zonename",
]


def check_and_parse_config(config: Dict[str, Any]) -> Dict[str, Any]:
    """Checks config, and returns a validated and defaults-filled config
    dictionary"""
    config = config.copy()
    unknown_options = set(config.keys()) - set(SUPPORTED_OPTIONS)
    if unknown_options:
        logger.warning("Unknown options ignored: %s", str(unknown_options))

    if config.get("verbose", False):
        config["loglevel"] = logging.INFO
    else:
        config["loglevel"] = logging.WARNING

    assert not (
        "input_dframe" not in config and "input_file" not in config
    ), "Specify either input_file or input_dframe"

    if "input_dframe" in config and "input_file" in config:
        raise ValueError("Do not specify both input_dframe and input_file")

    required_input_dframe_columns = ["DATE", "MD", "WELL_NAME", "PRESSURE"]
    if "input_dframe" not in config and "input_file" in config:
        input_dframe = pd.read_csv(config["input_file"])
        input_dframe["DATE"] = pd.to_datetime(input_dframe["DATE"], dayfirst=True)
        config["input_dframe"] = input_dframe
    missing_columns = set(required_input_dframe_columns) - set(
        config["input_dframe"].columns
    )
    optional_columns = ["TVD", "NORTH", "EAST"]
    for col in optional_columns:
        if col not in config["input_dframe"]:
            config["input_dframe"][col] = np.nan

    if missing_columns:
        raise ValueError(f"Missing columns {str(missing_columns)} in input table")

    config["exportdir"] = config.get("exportdir", ".")
    assert Path(config["exportdir"]).is_dir(), "exportdir did not exist, create upfront"

    config["interpolation"] = config.get("interpolation", "cubic")
    assert config["interpolation"] in [
        "linear",
        "cubic",
    ], "Only linear and cubic interpolation is supported"

    if "alias_file" in config:
        assert "alias" not in config, "Do not provide both alias and alias_file"
        try:
            alias_dframe = pd.read_csv(
                config["alias_file"], index_col=config.get("rms_name", "RMS_WELL_NAME")
            )
            config["alias"] = alias_dframe.to_dict()[
                config.get("ecl_name", "ECLIPSE_WELL_NAME")
            ]
            logger.info("Parsed RMS to Eclipse aliases from %s", config["alias_file"])
        except ValueError as err:
            raise ValueError(
                "Could not load RMS-Eclipse well alias file. Check column names"
            ) from err
    else:
        config["alias"] = config.get("alias", {})

    config["welldatefile"] = config.get("welldatefile", "well_date_rft.txt")
    assert isinstance(config["welldatefile"], str), "welldatefile must be a string"

    if "welldatefile" not in config:
        config["welldatefile"] = "well_date_rft.txt"

    config["rft_prefix"] = config.get("rft_prefix", "")
    assert isinstance(config["rft_prefix"], str), "rft_prefix must be a string"

    config["trajectory_name"] = config.get("trajectory_name", "Drilled trajectory")
    assert isinstance(
        config["trajectory_name"], str
    ), "trajectory_name must be a string"

    return config


def get_well_coords(
    project, wellname: str, trajectory_name: str = "Drilled trajectory"
):
    """Extracts  well coordinates as a (wellpoints, 3) numpy array
    from a ROXAPI RMS project reference.

    Args:
        project: Roxapi project reference.
        wellname: Name of well as it exists in the RMS project
    """
    return (
        project.wells[wellname]
        .wellbore.trajectories[trajectory_name]
        .survey_point_series.get_measured_depths_and_points()
    )


def strictly_downward(coords: np.ndarray) -> bool:
    """Check if a well trajectory has absolutely no horizontal sections

    Args:
        coords: n x 4 array of coordinates, only
            z values in 4th column is used.

    Returns:
        True if there are no horizontals.
    """
    return bool(np.all(coords[:-1, 3] < coords[1:, 3]))


def interp_from_md(
    md_value: float, coords, interpolation: str = "cubic"
) -> Tuple[float, float, float]:
    """
    Function to interpolate East, North, TVD values of a well point
    defined by its name (wname) and its corresponding MD point (md_value).

    The interpolation of the well trajectory will be linear if linear = True,
    using cubic spline otherwise.

    The coords input array should consist of rows::

        md, x, y, z

    Returns:
        x, y and z along the wellpath at requested measured depth value.
    """
    if coords is None:
        raise ValueError("Can't interpolate with no well coordinates")
    if interpolation.lower() == "linear":
        logger.info("Interpolating (linear) East, North, TVD from MD")
        return (
            np.interp(md_value, coords[:, 0], coords[:, 1]),
            np.interp(md_value, coords[:, 0], coords[:, 2]),
            np.interp(md_value, coords[:, 0], coords[:, 3]),
        )
    if interpolation.lower() == "cubic":
        logger.info("Interpolating (cubic spline) East, North, TVD from MD")
        cs_x = CubicSpline(coords[:, 0], coords[:, 1])  # East
        cs_y = CubicSpline(coords[:, 0], coords[:, 2])  # North
        cs_z = CubicSpline(coords[:, 0], coords[:, 3])  # TVD
        return (float(cs_x(md_value)), float(cs_y(md_value)), float(cs_z(md_value)))
    raise ValueError(f"Non-supported interpolation method: {interpolation}")


def interp_from_xyz(
    xyz: Tuple[float, float, float],
    coords: np.ndarray,
    interpolation: str = "cubic",
) -> float:
    """Interpolate MD value of a well point defined by its name (wellname)
    and its corresponding East, North, TVD values (xyz tuple)

    The coords input array should consist of rows::

        md, x, y, z

    The interpolation of the TVD well trajectory will be used
    (with linear algorithm if linear = True, using cubic spline otherwise)
    if the well is strictly going downward (TVD series stricly increasing).

    If the well is horizontal or in a hook shape, a projection of the point
    on the well trajectory will be used, using cubic spline interpolation
    of the well trajectory points.

    Args:
        coords
        xyz: x, y and z for the coordinate where MD is requested.
        interpolation: Use either "cubic" (default) or "linear"

    Returns:
        Measured depth value at (x,y,z)
    """

    # Interpolate to get corresponding MD

    logger.info("Interpolating MD from East, North, TVD")
    if strictly_downward(coords):
        # Interpolation from TVD survey can be used
        if interpolation.lower() == "linear":
            md_value = np.interp(xyz[2], coords[:, 3], coords[:, 0])
        else:
            md_value = float(CubicSpline(coords[:, 3], coords[:, 0])(xyz[2]))
        logger.info("MD estimated on strictly downward wellpath: %s", str(md_value))
        return md_value

    # When the well is not strictly downward MD must be used as a
    # parametrization of the trajectory coordinates (East, North, TVD)

    # RFT pressure point X0(x0,y0,z0) defined by:
    #     x0 = obs['EAST']
    #     y0 = obs['NORTH']
    #     z0 = obs['TVD']

    # distance of point X0 to any point X(x,y,z) of the well trajectory is:
    #     dist(X,X0)  = sqrt[ (x-x0)**2 + (y-y0)**2 + (z-z0)**2 ]
    #     dist(MD,X0) = sqrt[ (f1(MD)-x0)**2 + (f2(MD)-y0)**2 + (f3(MD)-z0)**2 ]

    # We are looking for the point X, belonging to the well, which is the closest
    # from X0, so we want to find the value of MD for which minimizes dist(MD,X0).
    # This MD value is a root of the derivative of the distance. We drop the sqrt()
    # since this is equivalent to minimazing the square of the distance.
    #
    #     d dist(MD,X0) / d MD = 0
    #
    # <=> 2[ (f1(MD)-x0)*f1'(MD) + (f2(MD)-y0)*f2'(MD) + (f3(MD)-z0)*f3'(MD) ] = 0
    #
    # <=> (f1(MD)-x0)*f1'(MD) + (f2(MD)-y0)*f2'(MD) + (f3(MD)-z0)*f3'(MD) = 0
    #
    # using the notation f' for the derivative of function f
    #
    # This can be computed analytically when using polynomial interpolations for
    # f1, f2, f3 (having this function as a new polynom and finding its real roots),
    # but this method is quite unstable depending the degree of the polynom,
    # it may respect the survey points but being quite erratic in-between.
    #
    # Instead, we use cubic spline interpolations which provide a much smoother curve,
    # but we cannot define analytically the sum and multiplication of splines,
    # so we need to compute numerically the distance and find the minimum

    # x = f1(MD) for any point X(x,y,z) of the trajectory defined by MD (x = East)
    poly_x = CubicSpline(coords[:, 0], coords[:, 1])
    # y = f2(MD) for any point X(x,y,z) of the trajectory defined by MD (y = North)
    poly_y = CubicSpline(coords[:, 0], coords[:, 2])
    # z = f3(MD) for any point X(x,y,z) of the trajectory defined by MD (z = TVD)
    poly_z = CubicSpline(coords[:, 0], coords[:, 3])

    # compute the square of the distance every 5 cm MD (can take a couple of seconds)
    step = 0.05
    dist_min = np.inf
    closest_md = None
    for my_md in np.arange(0, coords[-1, 0], step):
        dist = (
            (poly_x(my_md) - xyz[0]) ** 2
            + (poly_y(my_md) - xyz[1]) ** 2
            + (poly_z(my_md) - xyz[2]) ** 2
        )
        if dist < dist_min:
            closest_md = my_md
            dist_min = dist
    dist_min = dist_min ** (0.5)
    if closest_md is not None:
        md_value = round(closest_md, 2)
    logger.info(
        "MD estimated on undulating wellpath: {} (mismatch = {:.4f} m)".format(
            md_value, dist_min
        )
    )

    # Alternative using polynoms
    # degree = 7
    # poly_x = np.polyfit(coord[:, 0], coord[:, 1], degree) # x = f1(MD)
    # poly_y = np.polyfit(coord[:, 0], coord[:, 2], degree) # y = f2(MD)
    # poly_z = np.polyfit(coord[:, 0], coord[:, 3], degree) # z = f3(MD)

    # pder_x = np.polyder(poly_x)             # derivative, f1'(MD)
    # pder_y = np.polyder(poly_y)             # derivative, f2'(MD)
    # pder_z = np.polyder(poly_z)             # derivative, f3'(MD)

    # poly_x[-1] = poly_x[-1] - x0            # f1(MD) := f1(MD)-x0
    # poly_y[-1] = poly_y[-1] - y0            # f2(MD) := f2(MD)-y0
    # poly_z[-1] = poly_z[-1] - z0            # f3(MD) := f3(MD)-z0

    # Derivative of the distance function
    # pder_dist  = np.polyadd(np.polyadd(np.polymul(poly_x, pder_x),
    #                                    np.polymul(poly_y, pder_y)),
    #                         np.polymul(poly_z, pder_z))

    # Minimum of the (squared) distance function is the real root of this polynom
    # roots = np.roots(pder_dist)
    # interp_md = roots[np.isreal(roots)].real

    # Only one root should be real
    # assert len(interp_md) == 1, ('0 or more than 1 point have been found ({}) to fit '
    #                              'the well trajectory for this RFT data point!'
    #                              .format(len(interp_md)))

    # md_value = interp_md[0]
    # print('   Methode 3: MD = {}'.format(md_value))

    return md_value


def ertobs_df_to_files(
    dframe: pd.DataFrame,
    exportdir: str = ".",
    welldatefile: str = "well_date_rft.txt",
    filename: str = "rft_ertobs.csv",
) -> None:
    """
    Exports data from a dataframe into ERT observations files.
    The input here is essentially the input dataframe, but where all "holes" in
    MD or XYZ have been filled.

    If ZONE is included, it will be added to the output.

    Gendata will read well_date_rft.txt and then each of the obs and txt files.

    Syntax of the output of the .txt files are determined by semeio:
    https://github.com/equinor/semeio/blob/master/semeio/jobs/rft/utility.py#L21

    semeio.jobs.rft.trajectory.load_from_file() will parse the ``.txt`` files.
    https://github.com/equinor/semeio/blob/master/semeio/jobs/rft/trajectory.py#L273

    This function produces such files::

      <WELL_NAME>.txt  # (for all wells)
      <WELL_NAME>_<REPORT_STEP>.obs
      well_date_rft.txt

    <wellname>.txt is a hardcoded filename pattern in GENDATA_RFT (semeio)
    well_date_rft.txt is supplied as configuration to GENDATA_RFT (semeio)

    The obs-files must match the GENERAL_OBSERVATION arguments in the
    ERT config.

    In the welldatefile, a choice is made by this function on how to
    enumerate the REPORTSTEPS parameter, which is to be given to GENDATA in ERT config.
    Here it will be constructed on the enumeration of DATE for each given well,
    meaning the number will be 1 for the first date for each well, and 2 on the
    second DATE pr. well etc.

    Args:
        dframe: Contains data for RFT observations, one
            observations pr. row
        exportdir: Path to directory where export is to happen. Must
            exists.
        welldatefile: Filename to write the "index" of observations to.
        filename: Filename for raw CSV data, for future use in GENDATA_RFT
    """

    # Dump directly to CSV, this is for future use:
    dframe.to_csv(Path(exportdir) / filename, index=False, header=True)
    logger.info("Written CSV to %s", Path(exportdir) / filename)

    dframe = dframe.copy()  # Since we will modify it.

    if "ZONE" not in dframe:
        dframe["ZONE"] = ""

    dframe["DATE"] = pd.to_datetime(dframe["DATE"], dayfirst=True)
    dframe["REPORT_STEP"] = (
        dframe.groupby("WELL_NAME")["DATE"]
        .rank(method="dense", ascending=True)
        .astype(int)
    )

    for wellname, wellframe in dframe.groupby("WELL_NAME"):
        traj_filename = Path(exportdir) / (wellname + ".txt")

        well_trajectory = wellframe[
            ["EAST", "NORTH", "MD", "TVD", "ZONE"]
        ].drop_duplicates()

        well_trajectory.to_csv(traj_filename, sep=" ", index=False, header=False)
        logger.info("Written trajectory data to %s", traj_filename)

        for wname_rep_step, well_rep_step_frame in wellframe.groupby("REPORT_STEP"):
            obs_filename = Path(exportdir) / (
                wellname + "_" + str(wname_rep_step) + ".obs"
            )

            # Left join with the well trajectory frame, to ensure we
            # can output -1 for RFT measurements for points that are only
            # active on a subset of the report steps (i.e. dates):
            obs_frame = pd.merge(well_trajectory, well_rep_step_frame, how="left")
            obs_frame["PRESSURE"] = obs_frame["PRESSURE"].fillna(value=-1)
            obs_frame["ERROR"] = obs_frame["ERROR"].fillna(value=0)
            obs_frame[["PRESSURE", "ERROR"]].to_csv(
                obs_filename,
                sep=" ",
                index=False,
                header=False,
            )
            logger.info("Written obs file to %s", obs_filename)

    dframe["DAY"] = dframe["DATE"].dt.strftime("%d")
    dframe["MONTH"] = dframe["DATE"].dt.strftime("%m")
    dframe["YEAR"] = dframe["DATE"].dt.strftime("%Y")
    dframe.groupby(["WELL_NAME", "DATE"])[
        ["WELL_NAME", "DAY", "MONTH", "YEAR", "REPORT_STEP"]
    ].first().to_csv(Path(exportdir) / welldatefile, sep=" ", index=False, header=False)
    logger.info("Written welldata file to %s", Path(exportdir) / welldatefile)


def fill_missing_md_xyz(
    dframe: pd.DataFrame,
    coords_pr_well: Dict[str, np.ndarray],
    interpolation: str = "cubic",
) -> pd.DataFrame:
    """
    Fill missing MD or XYZ values in incoming dataframe, interpolating
    in given well trajectories.

    Args:
        dframe: Must contain WELL_NAME, EAST, NORTH, MD, TVD
        coords_pr_well: One key for each WELL_NAME, pointing to a n by 3
            numpy matrix with well coordinates (as outputted by roxapi)
    """
    for row_idx, row in dframe.iterrows():
        if pd.isnull(row["TVD"]):
            if pd.isnull(row["MD"]):
                logger.error("Both TVD and MD is missing for well %s", row["WELL_NAME"])
            else:
                xyz = interp_from_md(
                    row["MD"],
                    coords_pr_well[row["WELL_NAME"]],
                    interpolation=interpolation,
                )
                dframe.at[row_idx, "EAST"] = xyz[0]
                dframe.at[row_idx, "NORTH"] = xyz[1]
                dframe.at[row_idx, "TVD"] = xyz[2]
        elif pd.isnull(row["MD"]):
            if any(
                (
                    pd.isnull(row["EAST"]),
                    pd.isnull(row["NORTH"]),
                    pd.isnull(row["TVD"]),
                )
            ):
                logger.error(
                    "EAST, NORTH and/or TVD is missing, can't compute MD for well %s",
                    row["WELL_NAME"],
                )
            else:
                xyz = (row["EAST"], row["NORTH"], row["TVD"])
                dframe.at[row_idx, "MD"] = interp_from_xyz(
                    xyz, coords_pr_well[row["WELL_NAME"]], interpolation=interpolation
                )
    return dframe


def main(config: Optional[Dict[str, Any]] = None) -> None:
    """The main function to be called from a RMS client python script.

    Args:
        config (dict): configuration as a dictionary.

    """
    if config is None:
        config = {}
    config = check_and_parse_config(config)
    if config is None:
        raise ValueError("Configuration is invalid")

    logger.setLevel(config["loglevel"])

    welldatefile = "well_date_rft.txt"

    dframe = config["input_dframe"].copy()
    dframe["WELL_NAME"] = config["rft_prefix"] + dframe["WELL_NAME"]

    no_value = dframe["PRESSURE"].replace("", np.nan).isnull()
    if no_value.any():
        logger.error("PRESSURE missing for some rows:")
        logger.error("\n %s", str(dframe[no_value]))
        raise ValueError("PRESSURE not provided")

    if "project" in config:
        grid_model = config["project"].grid_models[config["gridname"]]
        grid = grid_model.get_grid()
        coords_pr_well = {}
        for well in dframe["WELL_NAME"].unique():
            coords_pr_well[well] = get_well_coords(
                config["project"], well, trajectory_name=config["trajectory_name"]
            )
        dframe = fill_missing_md_xyz(dframe, coords_pr_well)

        dframe["rms_cell_index"] = dframe.apply(
            lambda df: grid.get_cells_at_points([df["EAST"], df["NORTH"], df["TVD"]]),
            axis=1,
        )
    else:
        assert not dframe["MD"].isnull().any() and not dframe["TVD"].isnull().any(), (
            "MD and TVD must be supplied for all points "
            "when there is no RMS project available."
        )

    if "ERROR" not in dframe:
        dframe["ERROR"] = np.nan
    no_error_rows = dframe["ERROR"].replace("", np.nan).isnull()
    if "absolute_error" in config:
        dframe.loc[no_error_rows, "ERROR"] = config["absolute_error"]
    elif "relative_error" in config:
        dframe.loc[no_error_rows, "ERROR"] = (
            config["relative_error"] * dframe["PRESSURE"]
        )
    else:
        if no_error_rows.any():
            logger.error("No ERROR given for some or all points.")
            logger.error(
                "Provide defaults or insert in data table input.\n%s",
                str(dframe[no_error_rows]),
            )
            raise ValueError("ERROR not provided")

    if "ZONE" in dframe and "project" in config:
        cells_outside_grid = dframe["rms_cell_index"].isnull()
        # WARNING: If there are nan's in the dataframe, the dtype of
        # these index columns changes to float

        if cells_outside_grid.any():
            logger.warning("RFT points outside the grid:")
            logger.warning("\n %s", str(dframe[cells_outside_grid]))

        dframe.loc[~cells_outside_grid, "rms_cell_zone_val"] = dframe.loc[
            ~cells_outside_grid, "rms_cell_index"
        ].apply(
            lambda cell_index: grid_model.properties[config["zonename"]].get_values()[
                int(cell_index)
            ]
        )
        # What happens when there is no code_names in the RMS project? Which
        # exception will RMS raise?
        dframe.loc[~cells_outside_grid, "rms_cell_zone_str"] = dframe.loc[
            ~cells_outside_grid, "rms_cell_zone_val"
        ].apply(
            lambda zone_value: grid_model.properties[config["zonename"]].code_names[
                int(zone_value)
            ]
        )

        # Determine points that in the grid are not in the zone provided
        # in the input CSV:
        cells_in_wrong_zone = dframe["rms_cell_zone_str"] != dframe["ZONE"]
        if cells_in_wrong_zone.any():
            logger.warning(
                "Some points are in a different zone in the RMS grid:\n %s",
                str(dframe[cells_in_wrong_zone]),
            )
    elif "ZONE" in dframe:
        logger.warning("Cannot verify zones when no RMS project is provided")

    # Translate all well names according to alias:
    dframe["WELL_NAME"] = dframe["WELL_NAME"].apply(lambda x: config["alias"].get(x, x))

    # Export *.obs, *.txt and well_date_rft.txt
    ertobs_df_to_files(dframe, exportdir=config["exportdir"], welldatefile=welldatefile)



"""Run tests in RMS.

Creates a tmp RMS project in given version which is used as fixture for all other Roxar
API dependent tests.

This requires a ROXAPI license, and to be ran in a "roxenvbash" environment; hence
the decorator "roxapilicense"

"""
from pathlib import Path
from os.path import isdir
import shutil
import numpy as np
import pytest

import xtgeo

try:
    import roxar
except ImportError:
    pass


# ======================================================================================
# settings to create RMS project!

TMPD = Path("TMP")
TMPD.mkdir(parents=True, exist_ok=True)

TPATH = Path("../xtgeo-testdata")
PROJNAME = "tmp_project_qcreset.rmsxxx"

PRJ = str(TMPD / PROJNAME)

CUBEDATA1 = TPATH / "cubes/reek/syntseis_20000101_seismic_depth_stack.segy"
CUBENAME1 = "synth1"

SURFTOPS1 = [
    TPATH / "surfaces/reek/1/topreek_rota.gri",
    TPATH / "surfaces/reek/1/midreek_rota.gri",
    TPATH / "surfaces/reek/1/lowreek_rota.gri",
]

SURFCAT1 = "DS_whatever"
SURFNAMES1 = ["TopReek", "MidReek", "BaseReek"]

GRIDDATA1 = TPATH / "3dgrids/reek/reek_sim_grid.roff"
PORODATA1 = TPATH / "3dgrids/reek/reek_sim_poro.roff"
ZONEDATA1 = TPATH / "3dgrids/reek/reek_sim_zone.roff"
GRIDNAME1 = "Simgrid"
PORONAME1 = "PORO"
ZONENAME1 = "Zone"

WELLSFOLDER1 = TPATH / "wells/reek/1"
WELLS1 = ["OP1_perf.w", "OP_2.w", "OP_6.w", "XP_with_repeat.w"]


@pytest.mark.skipunlessroxar
@pytest.fixture(name="create_project", scope="module", autouse=True)
def fixture_create_project():
    """Create a tmp RMS project for testing, populate with basic data.

    After the yield command, the teardown phase will remove the tmp RMS project.
    """
    prj1 = str(PRJ)

    print("\n******** Setup RMS project!\n")
    if isdir(prj1):
        print("Remove existing project! (1)")
        shutil.rmtree(prj1)

    project = roxar.Project.create()

    rox = xtgeo.RoxUtils(project)
    print("Roxar version is", rox.roxversion)
    print("RMS version is", rox.rmsversion(rox.roxversion))
    assert "1." in rox.roxversion

    for wfile in WELLS1:
        wobj = xtgeo.well_from_file(WELLSFOLDER1 / wfile)
        if "XP_with" in wfile:
            wobj.name = "OP2_w_repeat"

        wobj.to_roxar(project, wobj.name, logrun="log", trajectory="My trajectory")

    # populate with cube data
    cube = xtgeo.cube_from_file(CUBEDATA1)
    cube.to_roxar(project, CUBENAME1, domain="depth")

    # populate with surface data
    rox.create_horizons_category(SURFCAT1)
    for num, name in enumerate(SURFNAMES1):
        srf = xtgeo.surface_from_file(SURFTOPS1[num])
        project.horizons.create(name, roxar.HorizonType.interpreted)
        srf.to_roxar(project, name, SURFCAT1)

    # populate with grid and props
    grd = xtgeo.grid_from_file(GRIDDATA1)
    grd.to_roxar(project, GRIDNAME1)
    por = xtgeo.gridproperty_from_file(PORODATA1, name=PORONAME1)
    por.to_roxar(project, GRIDNAME1, PORONAME1)
    zon = xtgeo.gridproperty_from_file(ZONEDATA1, name=ZONENAME1)
    zon.values = zon.values.astype(np.uint8)
    zon.to_roxar(project, GRIDNAME1, ZONENAME1)

    # save project (both an initla version and a work version) and exit
    project.save_as(prj1)
    project.close()

    yield project

    print("\n******* Teardown RMS project!\n")

    if isdir(prj1):
        print("Remove existing project! (1)")
        shutil.rmtree(prj1)


@pytest.mark.skipunlessroxar
def test_qcreset():
    """Test qcreset metod in roxapi."""
    # ==================================================================================
    # pylint: disable=invalid-name
    from fmu.tools.rms import qcreset

    rox = xtgeo.RoxUtils(project=PRJ)

    SETUP1 = {
        "project": rox._project,
        "horizons": {
            "DS_whatever": ["TopReek", "MidReek"],
        },
        "grid_models": {
            "Simgrid": ["PORO"],
        },
        "value": 0.088,
    }

    SETUP2 = {
        "project": rox._project,
        "horizons": {
            "DS_whatever": ["TopReek", "MidReek"],
        },
        "grid_models": {
            "Simgrid": ["PORO"],
        },
    }

    qcreset.set_data_constant(SETUP1)

    topr = xtgeo.surface_from_roxar(rox._project, "TopReek", "DS_whatever")
    assert topr.values.mean() == pytest.approx(0.088)

    poro = xtgeo.gridproperty_from_roxar(rox._project, "Simgrid", "PORO")
    assert poro.values.mean() == pytest.approx(0.088)

    qcreset.set_data_empty(SETUP2)

    top = rox._project.horizons["TopReek"]["DS_whatever"]

    assert top.is_empty() is True



# type: ignore
""" Module for generating a set of TornadoPlots
for webviz on webviz Pages. """

import os.path
import copy
from collections import OrderedDict
import pandas as pd
import yaml

from fmu.tools.sensitivities import summarize_design
from fmu.tools.sensitivities import calc_tornadoinput, find_combinations
from fmu import ensemble

try:
    from webviz import SubMenu, Page
    from webviz.page_elements import TornadoPlot

    HAS_WEBVIZ_STATIC = True
except ImportError:
    HAS_WEBVIZ_STATIC = False

WEBVIZ_STATIC_DEPRECIATION_WARNING = """
  **** Depreciation warning ****
This tornado plot generator is using a version of webviz
that is being deprecated and will not be supported throughout 2020.

Check https://github.com/equinor/webviz-subsurface for sensitivity
visualization using the new framework.
"""


def yconfig(inputfile):
    """Read from YAML file."""
    with open(inputfile, "r") as stream:
        config = yaml.load(stream)
    print("Input config YAML file <{}> is read...".format(inputfile))
    return config


def yconfig_set_defaults(config):
    """Override the YAML config with defaults where missing input."""
    newconfig = copy.deepcopy(config)

    # some defaults if data is missing
    if "webvizname" not in newconfig["tornadooptions"]:
        newconfig["tornadooptions"]["webvizname"] = "webviz_tornado"

    if "reference" not in newconfig["tornadooptions"]:
        newconfig["tornadooptions"]["reference"] = "0"

    if "scale" not in newconfig["tornadooptions"]:
        newconfig["tornadooptions"]["scale"] = "absolute"

    if "cutbyref" not in newconfig["tornadooptions"]:
        if "cutbyseed" in newconfig["tornadooptions"]:
            print(
                "Warning: Use of cutbyseed will be phased out." "Use cutbyref instead"
            )
            newconfig["tornadooptions"]["cutbyref"] = newconfig["tornadooptions"][
                "cutbyseed"
            ]
        else:
            newconfig["tornadooptions"]["cutbyref"] = "No"

    if "designsheet" not in newconfig["design"]:
        newconfig["design"]["designsheet"] = "DesignSheet01"

    if "gatherresults" not in newconfig["results"]:
        newconfig["results"]["gatherresults"] = False

    if "printresultfile" not in newconfig["results"]:
        newconfig["results"]["printresult"] = False

    if "renamecolumns" not in newconfig["results"]:
        newconfig["results"]["renamecolumns"] = False

    if "menuprefix" not in newconfig["calculations"]:
        newconfig["calculations"]["menuprefix"] = ""

    return newconfig


def make_xlabel(ref_value, scale="percentage", reference="seed"):
    """Makes xlabel for tornadoplot from input"""
    label = scale.capitalize() + " change compared to ref value "
    label += "{:.2e}".format(ref_value) + ".   "
    if reference.isdigit():
        label += "   Reference is realisation " + str(reference)
    else:
        label += "   Reference is sensititivity " + reference
    return label


def gatherresults(config):
    """Gathers .csv files from different realisations into one"""
    ensemblepaths = config["results"]["ensemblepaths"]
    print("Gathering csv files from:")
    print(ensemblepaths)
    singleresultfile = config["results"]["singleresultfile"]
    print(singleresultfile)

    doe_ensemble = ensemble.ScratchEnsemble("doe_ensemble", ensemblepaths)
    results = doe_ensemble.load_csv(singleresultfile)

    if config["results"]["writeresultfile"]:
        outdir = config["results"]["exportdir"]
        outfilename = config["results"]["exportfilename"]
        if not os.path.exists(outdir):
            os.mkdir(outdir)
        fullname = os.path.join(outdir, outfilename)
        results.to_csv(fullname, index=False)

    return results


def add_webviz_tornadoplots(web, configfile):
    """
    Generating a set of TornadoPlots for webviz on webviz Pages

    Args:
        web: webportal where submenu should be added
        configfile(str): yaml configuration file for tornado calculations

    Returns:
        object: Set of webviz.Pages with tornado plots

    Example:
        >>> from fmu.tools.sensitivities import add_webviz_tornadoplots
        >>> from webviz import Webviz
        >>> html_foldername = './webportal_example'
        >>> title = 'Snorreberg'
        >>> web = Webviz(title)
        >>> configfile ='../input/config/config_filename.yaml'
        >>> add_webviz_tornadoplots(web, configfile)
        >>> web.write_html(html_foldername, overwrite=True, display=True)

    """
    if not HAS_WEBVIZ_STATIC:
        print("ERROR: Cannot make Tornado plots without Webviz static")
        return None
    print(WEBVIZ_STATIC_DEPRECIATION_WARNING)

    yamlfile = configfile

    # input
    inputfile = yconfig(yamlfile)
    config = yconfig_set_defaults(inputfile)

    # assign parameters from config
    designpath = config["design"]["designpath"]
    designname = config["design"]["designname"]
    designsheet = config["design"]["designsheet"]

    response = config["calculations"]["responses"]

    selections = OrderedDict(config["calculations"]["selections"])
    selectors = selections.keys()

    reference = str(config["tornadooptions"]["reference"])
    scale = config["tornadooptions"]["scale"]
    cutbyref = config["tornadooptions"]["cutbyref"]
    menuprefix = config["calculations"]["menuprefix"]

    # Option: results from single realisation csv files needs gathering first
    if config["results"]["gatherresults"]:
        results = gatherresults(config)

    # Default: concatenated result file already exists as csv
    if not config["results"]["gatherresults"]:
        resultpath = config["results"]["resultpath"]
        resultname = config["results"]["resultname"]
        resultfile = os.path.join(resultpath, resultname)
        results = pd.read_csv(resultfile)

    # Option: renaming of columns in resultfile, e.g. from CSV_EXPORT1
    if config["results"]["renamecolumns"]:
        results.rename(columns=config["results"]["renaming"], inplace=True)

    # Read design matrix and find realisation numbers for each sensitivity
    designname = os.path.join(designpath, designname)
    designtable = summarize_design(designname, designsheet)
    print("Summary of design matrix:")
    with pd.option_context("expand_frame_repr", False):
        print(designtable)

    # Find all combinations of selections to generate plots for
    comb = find_combinations(selections)

    # Grouping calculation of tornado plots per response.
    # One list of plots under each response
    for res in response:
        smn = SubMenu(menuprefix + " " + res)
        for cmb in comb:
            pagetitle = res
            for ksel in range(len(selections)):
                pagetitle += " ,  " + str(selections.keys()[ksel])
                pagetitle += "  =  " + str(cmb[ksel]).strip("[]")
            pge = Page(pagetitle)
            (tornadotable, ref_value) = calc_tornadoinput(
                designtable, results, res, selectors, cmb, reference, scale, cutbyref
            )
            print("For response and selections: ")
            print(pagetitle)
            print("reference average: ", ref_value)
            print(
                "Calculation settings: ",
                scale,
                ", Cut by reference: ",
                cutbyref,
                ",  Reference: ",
                reference,
            )
            with pd.option_context("expand_frame_repr", False):
                print(tornadotable)
            print("\n")
            # While waiting for webviz issues 84 and 85 to be solved
            # xlabel = make_xlabel(ref_value, scale, reference)
            # tornado_plot = \
            # tornado_plot = TornadoPlot(tornadotable, xlabel, pagetitle)
            tornado_plot = TornadoPlot(tornadotable)
            pge.add_content(tornado_plot)
            smn.add_page(pge)

        web.add(smn)



"""
This private module is used for some common functions
"""
import sys


class _QCCommon(object):
    """
    Common functions, like print_info()
    """

    def __init__(self):

        self._verbosity = 0  # verbosity level

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, level):
        if level in ("info", 1):
            self._verbosity = 1
        elif level in ("debug", 2):
            self._verbosity = 2

    def print_info(self, string):
        """Do print based on verbosity level >= 1"""
        if self._verbosity > 0:
            print("INFO  >>", string)

    def print_debug(self, string):
        """Do debug print based on verbosity level >= 2"""
        if self._verbosity > 1:
            print("DEBUG >>", string)

    @staticmethod
    def give_warn(string):
        """Give warning to user"""
        print("WARN  >>", string)

    @staticmethod
    def force_stop(string):
        """Give stop message to STDERR and stop process"""
        mode = sys.stderr
        print()
        print("!" * 70, file=mode)
        print("STOP! >>", string, file=mode)
        print("!" * 70, file=mode)
        print()

        sys.exit(string)



"""Testing statistical helper functions for the design matrix generator"""

import numbers

import pytest

from fmu.tools.sensitivities import design_distributions as dists

# pylint: disable=protected-access


def test_check_dist_params_normal():
    """Test normal dist param checker"""
    # First element in returned 2-tuple is True or False:
    assert not dists._check_dist_params_normal([])[0]
    assert not dists._check_dist_params_normal(())[0]

    assert not dists._check_dist_params_normal([0])[0]
    assert not dists._check_dist_params_normal([0, 0, 0])[0]
    assert not dists._check_dist_params_normal([0, 0, 0, 0, 0])[0]

    assert not dists._check_dist_params_normal(["mean", "mu"])[0]

    assert dists._check_dist_params_normal([0, 1])[0]
    assert not dists._check_dist_params_normal([0, -1])[0]
    assert dists._check_dist_params_normal([0, 0])[0]  # edge case

    # Truncated
    assert dists._check_dist_params_normal([0, 1, 0, 1])[0]


def test_check_dist_params_lognormal():
    """Test lognormal dist param checker"""
    assert not dists._check_dist_params_lognormal([])[0]
    assert not dists._check_dist_params_lognormal(())[0]

    assert not dists._check_dist_params_lognormal([0])[0]
    assert not dists._check_dist_params_lognormal([0, 0, 0])[0]

    assert not dists._check_dist_params_lognormal(["mean", "mu"])[0]

    assert dists._check_dist_params_lognormal([0, 1])[0]
    assert not dists._check_dist_params_lognormal([0, -1])[0]

    assert dists._check_dist_params_lognormal([0, 0])[0]  # edge case


def test_check_dist_params_uniform():
    """Test lognormal dist param checker"""
    assert not dists._check_dist_params_uniform([])[0]
    assert not dists._check_dist_params_uniform(())[0]

    assert not dists._check_dist_params_uniform([0])[0]
    assert not dists._check_dist_params_uniform([0, 0, 0])[0]

    assert not dists._check_dist_params_uniform(["mean", "mu"])[0]

    assert dists._check_dist_params_uniform([0, 1])[0]
    assert not dists._check_dist_params_uniform([0, -1])[0]

    assert dists._check_dist_params_uniform([0, 0])[0]  # edge case


def test_check_dist_params_triang():
    """Test triang dist param checker"""
    assert not dists._check_dist_params_triang([])[0]
    assert not dists._check_dist_params_triang(())[0]

    assert not dists._check_dist_params_triang([0])[0]
    assert not dists._check_dist_params_triang([0, 0])[0]
    assert not dists._check_dist_params_triang([0, 0, 0, 0])[0]

    assert not dists._check_dist_params_triang(["foo", "bar", 0])[0]

    assert dists._check_dist_params_triang([0, 1, 2])[0]
    assert not dists._check_dist_params_triang([0, -1, -4])[0]
    assert not dists._check_dist_params_triang([0, 1000, 999.99])[0]

    assert dists._check_dist_params_triang([0, 0, 0])[0]  # edge case


def test_check_dist_params_pert():
    """Test pert dist param checker"""
    assert not dists._check_dist_params_pert([])[0]
    assert not dists._check_dist_params_pert(())[0]

    assert not dists._check_dist_params_pert([0])[0]
    assert not dists._check_dist_params_pert([0, 0])[0]
    assert not dists._check_dist_params_pert([0, 0, 0, 0, 0])[0]

    assert not dists._check_dist_params_pert(["foo", "bar", 0])[0]

    assert dists._check_dist_params_pert([0, 1, 2])[0]
    assert not dists._check_dist_params_pert([0, -1, -4])[0]
    assert not dists._check_dist_params_pert([0, 1000, 999.99])[0]
    assert dists._check_dist_params_pert([0, 1000, 1000, 999.99])[0]

    assert dists._check_dist_params_pert([0, 0, 0])[0]  # edge case


def test_check_dist_params_logunif():
    """Test logunif dist param checker"""
    assert not dists._check_dist_params_logunif([])[0]
    assert not dists._check_dist_params_logunif(())[0]

    assert not dists._check_dist_params_logunif([0])[0]
    assert not dists._check_dist_params_logunif([0, 0, 0])[0]

    assert not dists._check_dist_params_logunif(["foo", "bar"])[0]

    assert not dists._check_dist_params_logunif([0, 1])[0]
    assert dists._check_dist_params_logunif([0.00001, 1])[0]
    assert dists._check_dist_params_logunif([0.0000001, 0.00001])[0]
    assert not dists._check_dist_params_logunif([0.00001, 0.0000001])[0]

    assert dists._check_dist_params_logunif([1, 1])[0]


def test_draw_values_normal():
    """Test drawing values"""
    values = dists.draw_values_normal([0, 1], 10)
    assert len(values) == 10
    assert all([isinstance(value, numbers.Number) for value in values])

    values = dists.draw_values_normal([0, 10, -1, 2], 50)
    assert all([-1 <= value <= 2 for value in values])


def test_draw_values_lognormal():
    """Test drawing lognormal values"""
    assert not dists.draw_values_lognormal([100, 10], 0).size
    values = dists.draw_values_lognormal([100, 10], 10)
    assert len(values) == 10
    assert all([isinstance(value, numbers.Number) for value in values])
    assert all([value > 0 for value in values])


def test_draw_values_uniform():
    """Test drawing uniform values"""
    assert not dists.draw_values_uniform([10, 100], 0).size

    values = dists.draw_values_uniform([10, 100], 20)
    assert len(values) == 20
    assert all([isinstance(value, numbers.Number) for value in values])
    assert all([10 <= value <= 100 for value in values])


def test_draw_values_triangular():
    """Test drawing triangular values"""
    assert not dists.draw_values_triangular([10, 100, 1000], 0).size
    with pytest.raises(ValueError):
        assert not dists.draw_values_triangular([10, 100, 1000], -1)

    with pytest.raises(TypeError):
        assert not dists.draw_values_triangular([10, 100, 1000], "somestring")

    values = dists.draw_values_triangular([10, 100, 1000], 15)
    assert len(values) == 15
    assert all([isinstance(value, numbers.Number) for value in values])
    assert all([10 <= value <= 1000 for value in values])


def test_draw_values_pert():
    """Test drawing pert values"""
    assert not dists.draw_values_pert([10, 50, 100], 0).size

    values = dists.draw_values_pert([10, 50, 100], 20)
    assert len(values) == 20
    assert all([isinstance(value, numbers.Number) for value in values])
    assert all([10 <= value <= 100 for value in values])


def test_draw_values_loguniform():
    """Test drawing loguniform values"""
    assert not dists.draw_values_uniform([10, 100], 0).size

    values = dists.draw_values_uniform([10, 100], 20)
    assert len(values) == 20
    assert all([isinstance(value, numbers.Number) for value in values])
    assert all([10 <= value <= 100 for value in values])


def test_sample_discrete():
    """Test sampling discrete values"""
    outcomes = ["foo", "bar.com"]

    # NB: return type from sample_discrete is different from the others,
    # it returns a 3-tuple with the values in the second element.
    values = dists.sample_discrete([",".join(outcomes)], 10)[1]
    assert all([value in outcomes for value in values])

    assert not dists.sample_discrete([",".join(outcomes)], 0)[1].size
    with pytest.raises(ValueError):
        # pylint: disable=expression-not-assigned
        dists.sample_discrete([",".join(outcomes)], -1)[1]

    assert "foo" not in dists.sample_discrete([",".join(outcomes), "0,1"], 10)[1]


def test_draw_values():
    """Test the wrapper function for drawing values"""

    values = dists.draw_values("unif", [0, 1], 10)
    assert len(values) == 10
    assert all([isinstance(value, numbers.Number) for value in values])
    assert all([0 <= value <= 1 for value in values])

    values = dists.draw_values("UNIF", [0, 1], 10)
    assert len(values) == 10

    values = dists.draw_values("unifORM", [0, 1], 10)
    assert len(values) == 10

    values = dists.draw_values("UnifORMgarbagestillworks", [0, 1], 10)
    assert len(values) == 10

    with pytest.raises(ValueError):
        dists.draw_values("non-existing-distribution", [0, 10], 100)

    values = dists.draw_values("NORMAL", [0, 1], 10)
    assert len(values) == 10

    values = dists.draw_values("LOGNORMAL", [0.1, 1], 10)
    assert len(values) == 10

    values = dists.draw_values("Pert", [0.1, 1, 10], 10)
    assert len(values) == 10

    values = dists.draw_values("triangular", [0.1, 1, 10], 10)
    assert len(values) == 10

    values = dists.draw_values("logunif", [0.1, 1], 10)
    assert len(values) == 10



"""Testing code for generation of design matrices"""

from pathlib import Path
from packaging import version

import pandas as pd

import pytest

from fmu.tools.sensitivities import DesignMatrix, excel2dict_design

TESTDATA = Path(__file__).parent / "data"


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_generate_onebyone(tmpdir):
    """Test generation of onebyone design"""

    inputfile = TESTDATA / "config/design_input_example1.xlsx"
    input_dict = excel2dict_design(inputfile)

    design = DesignMatrix()
    design.generate(input_dict)
    # Checking dimensions of design matrix
    assert design.designvalues.shape == (80, 10)

    # Write to disk and check some validity
    tmpdir.chdir()
    design.to_xlsx("designmatrix.xlsx")
    assert Path("designmatrix.xlsx").exists
    diskdesign = pd.read_excel("designmatrix.xlsx", engine="openpyxl")
    assert "REAL" in diskdesign
    assert "SENSNAME" in diskdesign
    assert "SENSCASE" in diskdesign
    assert not diskdesign.empty

    diskdefaults = pd.read_excel(
        "designmatrix.xlsx", sheet_name="DefaultValues", engine="openpyxl"
    )
    assert not diskdefaults.empty
    assert len(diskdefaults.columns) == 2


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_generate_full_mc(tmpdir):
    """Test generation of full monte carlo"""
    inputfile = TESTDATA / "config/design_input_mc_with_correls.xlsx"
    input_dict = excel2dict_design(inputfile)

    design = DesignMatrix()
    design.generate(input_dict)

    # Checking dimensions of design matrix
    assert design.designvalues.shape == (500, 16)

    # Checking reproducibility from distribution_seed
    assert design.designvalues["PARAM1"].sum() == 17.419

    # Write to disk and check some validity
    tmpdir.chdir()
    design.to_xlsx("designmatrix.xlsx")
    assert Path("designmatrix.xlsx").exists
    diskdesign = pd.read_excel(
        "designmatrix.xlsx", sheet_name="DesignSheet01", engine="openpyxl"
    )
    assert "REAL" in diskdesign
    assert "SENSNAME" in diskdesign
    assert "SENSCASE" in diskdesign
    assert not diskdesign.empty

    diskdefaults = pd.read_excel(
        "designmatrix.xlsx", sheet_name="DefaultValues", engine="openpyxl"
    )
    assert not diskdefaults.empty
    assert len(diskdefaults.columns) == 2



"""The qcproperties module"""
from pathlib import Path
from typing import Optional, Any
import pandas as pd
import yaml

from fmu.tools._common import _QCCommon
from fmu.tools.qcdata import QCData

from fmu.tools.qcproperties._grid2df import GridProps2df
from fmu.tools.qcproperties._well2df import WellLogs2df
from fmu.tools.qcproperties._aggregate_df import PropertyAggregation

QCC = _QCCommon()


class QCProperties:
    """
    The QCProperties class consists of a set of methods for extracting
    property statistics from 3D Grids, Raw and Blocked wells.

    Statistics can be collected from either discrete or continous properties.
    Dependent on the property different statistics are collected.

    The methods for statistics extraction can be run individually, or a
    yaml-configuration file can be used to enable an automatic run of the
    methods. See the method 'from_yaml'.

    When several methods of statistics extraction has been run within the instance,
    a merged dataframe is available through the 'dataframe' property.

    All methods can be run from either RMS python, or from files.

    XTGeo is being utilized to get a dataframe from the input parameter data.
    XTGeo data is reused in the instance to increase performance.

    """

    def __init__(self):

        self._xtgdata = QCData()  # QCData instance, general XTGeo data
        self._dfs = []  # list of dataframes with aggregated statistics
        self._selectors_all = []
        self._proptypes_all = []
        self._ids = []
        self._dataframe = pd.DataFrame()  # merged dataframe with statistics

    # Properties:
    # ==================================================================================

    @property
    def dataframe(self):
        """Dataframe with statistics"""
        self._dataframe = self._create_or_return_dataframe()
        return self._dataframe

    # Hidden methods:
    # ==================================================================================

    def _initiate_from_config(self, cfg: str, project: Optional[object]):
        """Run methods for statistics extraction based on entries in yaml-config"""
        with open(cfg, "r") as stream:
            data = yaml.safe_load(stream)

        if "grid" in data:
            for item in data["grid"]:
                self.get_grid_statistics(data=item, project=project)

        if "wells" in data:
            for item in data["wells"]:
                self.get_well_statistics(data=item, project=project)

        if "blockedwells" in data:
            for item in data["blockedwells"]:
                self.get_bwell_statistics(data=item, project=project)

    def _create_or_return_dataframe(self):
        """
        Combine dataframes from all runs within the instance.
        Only update dataframe if more data have been run within the
        instance, else return previous dataframe.
        """
        dframe = self._dataframe
        dframes = self._dfs

        if dframe.empty or len(dframes) > len(dframe["ID"].unique()):
            QCC.print_debug("Updating combined dataframe")
            self._warn_if_different_property_types()
            dframe = pd.concat(dframes)

            # fill NaN with "Total" for dataframes with missing selectors
            dframe[self._selectors_all] = dframe[self._selectors_all].fillna("Total")

            # Specify column order in statistics dataframe
            cols_first = ["PROPERTY"] + self._selectors_all
            dframe = dframe[
                cols_first + [x for x in dframe.columns if x not in cols_first]
            ]
        return dframe

    def _warn_if_different_property_types(self):
        """Give warning if dataframes have different property types"""
        if not all(ptype == self._proptypes_all[0] for ptype in self._proptypes_all):
            QCC.give_warn(
                "Merging statistics dataframes from different property types "
                "(continous/discrete). Is this intentional?"
            )

    def _adjust_id_if_duplicate(self, run_id: str) -> str:
        """
        Check for equal run ids, modify ids
        by adding a number to get them unique.
        """
        check_id = run_id
        count = 0
        while check_id in self._ids:
            check_id = f"{run_id}({count+1})"
            count += 1
        return check_id

    def _set_dataframe_id_and_class_attributes(
        self, statistics: PropertyAggregation, source: str, run_id: str
    ):
        """
        Set source and id column of statistics datframe, and different
        class attributes.
        """
        run_id = self._adjust_id_if_duplicate(run_id)
        # set id and source columns in statistics dataframe
        statistics.dataframe["ID"] = run_id
        statistics.dataframe["SOURCE"] = source

        self._ids.append(run_id)
        self._dfs.append(statistics.dataframe)

        for selector in statistics.controls["selectors"]:
            if selector not in self._selectors_all:
                self._selectors_all.append(selector)

        self._proptypes_all.append(statistics.controls["property_type"])

    # pylint: disable = no-self-argument, not-callable
    def _check_multiple_filters(method: Any):
        """Decorator function for extracting statistics with different filters"""

        def wrapper(self, **kwargs):
            if "multiple_filters" in kwargs["data"]:
                for name, filters in kwargs["data"]["multiple_filters"].items():
                    kwargs["data"].update(filters=filters, name=name)
                    method(self, **kwargs)
                return self.dataframe
            return method(self, **kwargs)

        return wrapper

    @_check_multiple_filters
    def _extract_statistics(
        self, dtype: str, data: dict, project: Optional[object], source: str
    ):
        """Create dataframe from properties and extract statistics"""
        QCC.verbosity = data.get("verbosity", 0)
        QCC.print_info("Starting run...")

        # Create Property dataframe from input (using XTGeo)
        property_data = (
            GridProps2df(project=project, data=data, xtgdata=self._xtgdata)
            if dtype == "grid"
            else WellLogs2df(
                project=project,
                data=data,
                xtgdata=self._xtgdata,
                blockedwells=dtype == "bwells",
            )
        )

        # Compute statistics
        stats = PropertyAggregation(property_data)

        self._set_dataframe_id_and_class_attributes(
            stats,
            source=source,
            run_id=data.get("name", source),
        )

        return stats.dataframe

    # QC methods:
    # ==================================================================================

    def get_grid_statistics(
        self,
        data: dict,
        project: Optional[object] = None,
    ) -> pd.DataFrame:
        """Extract property statistics from 3D Grid"""
        return self._extract_statistics(
            dtype="grid",
            data=data,
            project=project,
            source=data.get("source", Path(data["grid"]).stem),
        )

    def get_well_statistics(
        self,
        data: dict,
        project: Optional[object] = None,
    ) -> pd.DataFrame:
        """Extract property statistics from wells"""
        return self._extract_statistics(
            dtype="wells",
            data=data,
            project=project,
            source=data.get("source", "wells"),
        )

    def get_bwell_statistics(
        self,
        data: dict,
        project: Optional[object] = None,
    ) -> pd.DataFrame:
        """Extract property statistics from blocked wells"""
        return self._extract_statistics(
            dtype="bwells",
            data=data,
            project=project,
            source=data.get(
                "source",
                "bwells" if project is None else data["wells"].get("bwname", "BW"),
            ),
        )

    def from_yaml(self, cfg: str, project: Optional[object] = None):
        """Use yaml-configuration file to run the statistics extractions methods"""
        self._initiate_from_config(cfg, project)

    def to_csv(self, csvfile: str):
        """Write combined dataframe to csv"""
        self.dataframe.to_csv(csvfile, index=False)
        QCC.print_info(f"Dataframe with statistics written to {csvfile}")



from packaging import version

import pytest

from fmu.tools.fipmapper import fipmapper

import pandas as pd


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.2"),
    reason="Pandas 0.25.2 is required for fipmappers disjointsets()",
)
@pytest.mark.parametrize(
    "mapdata, expected_disjointsets",
    [
        (
            # The first test proves the code on a three-by-three reservoir model
            # which includes the major "cruxes" to be solved. This serves
            # as both the motivation for the code and as demonstration of
            # the solution.
            #
            # In the geomodel, the reservoir is three-by-three
            # with region-zone pairs as follows:
            #
            #    +--------+--------+--------+
            #    | (A, U) | (B, U) | (C, U) |
            #    +--------+--------+--------+
            #    | (A, M) | (B, M) | (C, M) |
            #    +--------+--------+--------+
            #    | (A, L) | (B, L) | (C, L) |
            #    +--------+--------+--------+
            #
            # while in the dynamical (Eclipse) model the reservoir has
            # a FIPNUM split like this:
            #
            #  regs: A    |   B     |    C       zones:
            #     +-----------------+----+----+
            #     |                 |  2 |  3 |    U
            #     |       1         +----+----+
            #     |                 |         |    M
            #     +-----------------+    4    |
            #     |       5         |         |    L
            #     +-----------------+---------+
            #
            # The problem the code solves is to determine which of these cells
            # that must be joined in order to have comparable numbers (when say
            # volume is available for each cell in the geomodel and dynamical
            # model separately.
            #
            # In this case, the joined structure must be a 4-cell partition:
            #
            #     +-----------------+---------+
            #     |                 |    3    |
            #     |       1         +---------+
            #     |                 |         |
            #     +-----------------+    2    |
            #     |       0         |         |
            #     +-----------------+---------+
            #
            # where the code identifies each set of joined cells by a
            # unique integer from a consecutive list starting at zero.
            # The enumeration of sets are predictable, based on lexiographical
            # sorts of REGION, ZONE and FIPNUM.
            {
                # The input is just mappings from the (region, zone)
                # "tuple" in the geomodel into FIPNUM in the dynamical
                # model. The spatial layout of regions, zones and FIPNUM
                # as displayed above does not matter.
                "region2fipnum": {
                    "A": [1, 5],
                    "B": [1, 5],
                    "C": [2, 3, 4],
                },
                "zone2fipnum": {
                    "U": [1, 2, 3],
                    "M": [1, 4],
                    "L": [5, 4],
                },
            },
            pd.DataFrame(
                columns=["REGION", "ZONE", "FIPNUM", "SET"],
                data=[
                    ["A", "L", 5, 0],
                    ["A", "M", 1, 1],
                    ["A", "U", 1, 1],
                    ["B", "L", 5, 0],
                    ["B", "M", 1, 1],
                    ["B", "U", 1, 1],
                    ["C", "L", 4, 2],
                    ["C", "M", 4, 2],
                    ["C", "U", 2, 3],
                    ["C", "U", 3, 3],
                ],
            ),
        ),
        pytest.param({}, [], marks=pytest.mark.xfail(raises=AssertionError)),
        (
            # Only one cell in both regzone partition and fipnum partition:
            {"region2fipnum": {"A": 1}, "zone2fipnum": {"U": 1}},
            [{"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0}],
        ),
        (
            # Same as above, but test that we can provide it as list input:
            {"region2fipnum": {"A": [1]}, "zone2fipnum": {"U": [1]}},
            [{"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0}],
        ),
        (
            # Same as above, but with integer regions and zone:
            {"region2fipnum": {1: 1}, "zone2fipnum": {1: 1}},
            [{"REGION": "1", "ZONE": "1", "FIPNUM": 1, "SET": 0}],
        ),
        (
            # FIPNUM split in two, gives only one group in return:
            {"region2fipnum": {"A": [1, 2]}, "zone2fipnum": {"U": [1, 2]}},
            [
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "A", "ZONE": "U", "FIPNUM": 2, "SET": 0},
            ],
        ),
        (
            # FIPNUM split in two, and two zones, gives two groups in return:
            {"region2fipnum": {"A": [1, 2]}, "zone2fipnum": {"U": [1], "L": [2]}},
            [
                {"REGION": "A", "ZONE": "L", "FIPNUM": 2, "SET": 0},
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 1},
            ],
        ),
        (
            # Two zones, one FIPNUM, gives one group in return:
            {"region2fipnum": {"A": 1}, "zone2fipnum": {"U": [1], "L": [1]}},
            [
                {"REGION": "A", "ZONE": "L", "FIPNUM": 1, "SET": 0},
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
            ],
        ),
        (
            # Two regions, one FIPNUM, gives one group in return:
            {"region2fipnum": {"A": 1, "B": 1}, "zone2fipnum": {"U": 1}},
            [
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 1, "SET": 0},
            ],
        ),
        (
            # Two regions, two zones, 4 FIPNUMs -> 4 groups:
            {
                "region2fipnum": {"A": [2, 1], "B": [3, 4]},
                "zone2fipnum": {"U": [1, 3], "L": [2, 4]},
            },
            [
                {"REGION": "A", "ZONE": "L", "FIPNUM": 2, "SET": 0},
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 1},
                {"REGION": "B", "ZONE": "L", "FIPNUM": 4, "SET": 2},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 3, "SET": 3},
            ],
        ),
        (
            # Two regions, two zones, 2 horizontal FIPNUMs -> 2 groups:
            {
                "region2fipnum": {"A": [2, 1], "B": [1, 2]},
                "zone2fipnum": {"U": 1, "L": [2, 2]},  # Double for L is ignored.
            },
            [
                {"REGION": "A", "ZONE": "L", "FIPNUM": 2, "SET": 0},
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 1},
                {"REGION": "B", "ZONE": "L", "FIPNUM": 2, "SET": 0},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 1, "SET": 1},
            ],
        ),
        (
            # Two regions, two zones, 2 region-wise FIPNUMs -> 2 groups:
            {
                "region2fipnum": {"A": 1, "B": 2},
                "zone2fipnum": {"U": [1, 2], "L": [1, 2]},
            },
            [
                {"REGION": "A", "ZONE": "L", "FIPNUM": 1, "SET": 0},
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "B", "ZONE": "L", "FIPNUM": 2, "SET": 1},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 2, "SET": 1},
            ],
        ),
        (
            # Two regions, one zone, 2 horizontal FIPNUMs which don't align
            # with the region boundary -> 1 group:
            {
                "region2fipnum": {"A": [1], "B": [1, 2]},
                "zone2fipnum": {"U": [1, 2]},
            },
            [
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 2, "SET": 0},
            ],
        ),
        (
            # Integer datatypes for regions and zones, always
            # strings in returned dataframe, even though
            # FipMapper handles the integers as ints internally
            {
                "region2fipnum": {1: 1, 2: 2},
                "zone2fipnum": {1: [1, 2], 2: [1, 2]},
            },
            [
                {"REGION": "1", "ZONE": "1", "FIPNUM": 1, "SET": 0},
                {"REGION": "1", "ZONE": "2", "FIPNUM": 1, "SET": 0},
                {"REGION": "2", "ZONE": "1", "FIPNUM": 2, "SET": 1},
                {"REGION": "2", "ZONE": "2", "FIPNUM": 2, "SET": 1},
            ],
        ),
        (
            # Mixed datatype int/str for regions and zones
            {
                "region2fipnum": {1: 1, "B": 2},
                "zone2fipnum": {"U": [1, 2], 2: [1, 2]},
            },
            [
                {"REGION": "1", "ZONE": "2", "FIPNUM": 1, "SET": 0},
                {"REGION": "1", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "B", "ZONE": "2", "FIPNUM": 2, "SET": 1},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 2, "SET": 1},
            ],
        ),
    ],
)
def test_disjointsets(mapdata, expected_disjointsets):
    expected_dframe = pd.DataFrame(expected_disjointsets)
    cols = list(expected_dframe.columns)
    mapper = fipmapper.FipMapper(mapdata=mapdata)
    pd.testing.assert_frame_equal(
        mapper.disjoint_sets()[cols].sort_values(by=cols, axis=0),
        expected_dframe.sort_values(by=cols, axis=0),
    )


@pytest.mark.parametrize(
    "dframe_records, expected_regions, expected_zones, "
    "expected_fipnums, expected_regzonefips",
    [
        ([{}], {}, {}, {}, {}),
        (
            [{"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0}],
            {0: ["A"]},
            {0: ["U"]},
            {0: [1]},
            {0: [("A", "U", 1)]},
        ),
        (
            [
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 1, "SET": 0},
            ],
            {0: ["A", "B"]},
            {0: ["U"]},
            {0: [1]},
            {0: [("A", "U", 1), ("B", "U", 1)]},
        ),
        (
            [
                # Switched order of rows compared to above:
                {"REGION": "B", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
            ],
            {0: ["A", "B"]},
            {0: ["U"]},
            {0: [1]},
            {0: [("A", "U", 1), ("B", "U", 1)]},
        ),
        (
            [
                {"REGION": "A", "ZONE": "U", "FIPNUM": 1, "SET": 0},
                {"REGION": "B", "ZONE": "U", "FIPNUM": 2, "SET": 1},
            ],
            {0: ["A"], 1: ["B"]},
            {0: ["U"], 1: ["U"]},
            {0: [1], 1: [2]},
            {0: [("A", "U", 1)], 1: [("B", "U", 2)]},
        ),
        (
            # The full example from the top of this file:
            pd.DataFrame(
                columns=["REGION", "ZONE", "FIPNUM", "SET"],
                data=[
                    ["A", "L", 5, 0],
                    ["A", "M", 1, 1],
                    ["A", "U", 1, 1],
                    ["B", "L", 5, 0],
                    ["B", "M", 1, 1],
                    ["B", "U", 1, 1],
                    ["C", "L", 4, 2],
                    ["C", "M", 4, 2],
                    ["C", "U", 2, 3],
                    ["C", "U", 3, 3],
                ],
            ).to_dict(orient="records"),
            {0: ["A", "B"], 1: ["A", "B"], 2: ["C"], 3: ["C"]},
            {0: ["L"], 1: ["M", "U"], 2: ["L", "M"], 3: ["U"]},
            {0: [5], 1: [1], 2: [4], 3: [2, 3]},
            {
                0: [("A", "L", 5), ("B", "L", 5)],
                1: [("A", "M", 1), ("A", "U", 1), ("B", "M", 1), ("B", "U", 1)],
                2: [("C", "L", 4), ("C", "M", 4)],
                3: [("C", "U", 2), ("C", "U", 3)],
            },
        ),
    ],
)
def test_set_lookups(
    dframe_records,
    expected_regions,
    expected_zones,
    expected_fipnums,
    expected_regzonefips,
):
    dframe = pd.DataFrame(dframe_records)
    assert fipmapper.regions_in_set(dframe) == expected_regions
    assert fipmapper.zones_in_set(dframe) == expected_zones
    assert fipmapper.fipnums_in_set(dframe) == expected_fipnums
    assert fipmapper.regzonefips_in_set(dframe) == expected_regzonefips



import os
import pytest


def pytest_runtest_setup(item):
    """Called for each test, see also pytest section in setup.cfg."""

    markers = [value.name for value in item.iter_markers()]

    # pytest.mark.skipifroxar:
    if "skipifroxar" in markers:
        if "ROXENV" in os.environ:
            pytest.skip("Skip test in ROXENV (env variable ROXENV is present)")

    # pytest.mark.skipunlessroxar:
    if "skipunlessroxar" in markers:
        if "ROXENV" not in os.environ:
            pytest.skip("Skip test if outside ROXENV (env variable ROXENV is present)")



from os.path import abspath
import pytest


@pytest.fixture()
def data_grid():
    return {
        "path": abspath("../xtgeo-testdata/3dgrids/reek/"),
        "grid": "reek_sim_grid.roff",
        "properties": {
            "PORO": {"name": "reek_sim_poro.roff"},
            "PERM": {"name": "reek_sim_permx.roff"},
        },
        "selectors": {
            "ZONE": {"name": "reek_sim_zone.roff"},
            "FACIES": {"name": "reek_sim_facies2.roff"},
        },
        "verbosity": 1,
    }


@pytest.fixture()
def data_wells():
    return {
        "path": abspath("../xtgeo-testdata/wells/reek/1/"),
        "wells": ["OP_*.w"],
        "properties": {
            "PORO": {"name": "Poro"},
            "PERM": {"name": "Perm"},
        },
        "selectors": {
            "ZONE": {"name": "Zonelog"},
            "FACIES": {"name": "Facies"},
        },
        "verbosity": 1,
    }


@pytest.fixture()
def data_bwells():
    return {
        "path": abspath("../xtgeo-testdata/wells/reek/1/"),
        "wells": ["OP_1.bw"],
        "properties": {
            "PORO": {"name": "Poro"},
        },
        "selectors": {
            "FACIES": {"name": "Facies"},
        },
        "verbosity": 1,
    }



"""The _qcforward module contains the base class"""

import sys
from copy import deepcopy
from os.path import join

import pandas as pd
import yaml
from fmu.tools._common import _QCCommon
from fmu.tools.qcdata import QCData

QCC = _QCCommon()


class QCForward:
    """
    The QCforward base class which has a set of QC functions that can be ran from
    either RMS python, or on disk. The input `data` will be
    somewhat different for the two run environents.

    It should be easy to add new functions to this class. The idea is to reuse
    as much as possible, and principles are:

    * For the client (user), the calling scripts shall be lean

    * All methods shall have a rich documention with examples, i.e. it shall
      be possible for users with less skills in scripting to copy/paste and then modify
      to their needs.

    """

    def __init__(self):
        self._method = None
        self._data = None  # input data dictionary
        self._path = "."
        self._gdata = QCData()  # QCData instance, stores XTGeo data
        self._ldata = None  # special data instance, for local data parsed per method
        self._reports = []  # List of all report, used to determin write/append mode

    @property
    def reports(self):
        return self._reports

    @reports.setter
    def reports(self, data):
        self._reports = data

    @property
    def data(self):
        """The input data dictionary"""
        return self._data

    @property
    def gdata(self):
        """A QCData() instance"""
        return self._gdata

    @gdata.setter
    def gdata(self, obj):
        """A QCData() instance"""
        if not isinstance(obj, QCData):
            raise TypeError("Wrong type of object, shall be QCData")
        self._gdata = obj

    @property
    def ldata(self):
        return self._ldata

    @ldata.setter
    def ldata(self, data):
        self._ldata = data

    def handle_data(self, data, project):

        data_is_yaml = True

        # data may be a yaml file
        if isinstance(data, str):
            try:
                with open(data, "r") as stream:
                    xdata = yaml.safe_load(stream)
            except FileNotFoundError as err:
                raise RuntimeError from err
            data_is_yaml = False
        else:
            xdata = data.copy()

        QCC.verbosity = xdata.get("verbosity", None)

        if data_is_yaml and "dump_yaml" in xdata and xdata["dump_yaml"]:
            xdata.pop("dump_yaml", None)
            with open(join(self._path, data["dump_yaml"]), "w") as stream:
                yaml.safe_dump(
                    xdata,
                    stream,
                    default_flow_style=None,
                )
            QCC.print_info("Dumped YAML to {}".format(data["dump_yaml"]))

        xdata["project"] = None
        if project:
            xdata["project"] = project
            QCC.print_info("Project type is {}".format(type(project)))

        return xdata

    def make_report(
        self, results: dict, reportfile: str = None, nametag: str = None
    ) -> pd.DataFrame:
        """Make a report which e.g. can be used in webviz plotting"""

        dfr = pd.DataFrame(results).assign(NAMETAG=nametag)

        if reportfile is not None:
            if reportfile in self.reports:
                dfr.to_csv(reportfile, index=False, mode="a", header=None)
            else:
                dfr.to_csv(reportfile, index=False)
                self._reports.append(reportfile)

        return dfr

    def evaluate_qcreport(self, dfr, name, stopaction=True) -> str:
        """Evalute and do actions on dataframe which contains the gridquality report.

        Action is done here if keyword action is True; otherwise the caller must decide
        upon action dependent on return status.

        Args:
            dfr (DataFrame): Pandas dataframe which needs a STATUS column with "OK",
                "WARN" or "STOP"
            name (str): Name of feature is under evaluation, e.g. "grid
                quality"
            stopaction(bool): If True and a STOP status is found, then a force_stop
                is done here, otherwise the status is returned and the caller need
                to to care of any actions.

        Returns:
            The string "STOP" or "CONTINUE" unless action is True and STOP is found
            and the result dataframe
        """
        statuslist = ("OK", "WARN", "STOP")

        dfr_status = None
        for status in statuslist:
            dfr_status = dfr[dfr["STATUS"] == status]
            if len(dfr_status) > 0:
                print(f"Status {status} for <{name}> nametag: {self.ldata.nametag})")

                stream = sys.stderr if status == "STOP" else sys.stdout

                dfr_status_print = dfr_status.to_string()
                print(f"{dfr_status_print}\n", file=stream)
                if status == "STOP":
                    if stopaction:
                        QCC.force_stop("STOP criteria is found!")
                    return status
        print(
            f"\n== QC forward check {self.__class__.__name__} "
            f"{self.ldata.nametag}) finished =="
        )

        return "CONTINUE"


class ActionsParser:
    def __init__(self, rule, mode="warn", verbosity="info"):

        QCC.verbosity = verbosity
        self.status = None  # in case no rule is set
        self.all = True
        self.compare = ">"
        self.limit = 90
        self.given = "<"
        self.criteria = 50
        self.mode = mode
        self.expression = "UNDEF"

        QCC.print_debug(f"Parse action: {rule}")

        if rule:
            self.parse(rule)

    def parse(self, rule):
        """Parse a rule given for an action.

        Args:
            rule (str): An expression on form "eachwell < 90%" or
                "allcells < 80% when > 99"

        """

        items = rule.split()
        self.status = "rule"
        self.expression = rule.replace(" ", "")

        if len(items) == 6:
            # 'all > 3% when < 20'
            self.all = "all" in items[0]
            self.compare = items[1]
            self.limit = float(items[2].replace("%", ""))
            self.given = items[4]
            self.criteria = float(items[5])

        elif len(items) == 3:
            self.all = "all" in items[0]
            self.compare = items[1]
            self.limit = float(items[2].replace("%", ""))
            self.given = ""
            self.criteria = ""

        else:
            raise ValueError(f"Input is wrong somehow: {rule}")

        key = "all" if self.all else "any"
        self.expression = key + self.compare + str(self.limit) + "%"
        if self.given:
            self.expression += "ifx" + self.given + str(self.criteria)


def actions_validator(actionsin: dict) -> dict:
    """General function to validate that the 'actions' input is on the correct form.

    The typical form is::

        "actions": [
            {"warn": "anywell < 80%", "stop": "anywell < 75%"},
            {"warn": "allwells < 90%", "stop": "allwells < 85%"},
        ],

    Several checks can be implemented; currently many is in a to-do state.

    Args:
        actionsin: Input actions record

    Returns:
        An actions python dictionary (potentially processed)
    """

    actions = deepcopy(actionsin)

    if not isinstance(actions, list):
        raise ValueError("The actions input must be a list")

    if len(actions) != 2:
        raise ValueError("The actions input must be a list with two element")

    # check each line to see if the rules seems logical
    anyhit = False
    allhit = False
    for elem in actions:
        if not set(elem.keys()).issubset(["warn", "stop"]):
            raise ValueError(
                "Both criteria 'warn' and 'stop' are required in actions! "
                f"You have keys: {list(elem.keys())}"
            )
        vals = list(elem.values())
        if "any" in str(vals) and "all" in str(vals):
            # avoid that one entry mixes all and any
            raise ValueError(
                "Seems wrong to mix 'all' and 'any' in one line which is not allowed"
            )
        if "any" in str(vals):
            anyhit = True
        if "all" in str(vals):
            allhit = True

    if not (anyhit and allhit):
        raise ValueError(
            f"One of 'any' or 'all' is missing in the actions input: {actions}"
        )

    return actions



"""Testing excel2dict"""
import os
from packaging import version

import pytest

import numpy as np
import pandas as pd

from fmu.tools.sensitivities import excel2dict_design, inputdict_to_yaml
from fmu.tools.sensitivities._excel2dict import _has_value

MOCK_GENERAL_INPUT = pd.DataFrame(
    data=[
        ["designtype", "onebyone"],
        ["repeats", "10"],
        ["rms_seeds", "default"],
        ["background", "None"],
        ["distribution_seed", "None"],
    ]
)

MOCK_DESIGNINPUT = pd.DataFrame(
    data=[["sensname", "numreal", "type", "param_name"], ["rms_seed", "", "seed"]]
)


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_excel2dict_design(tmpdir):
    """Test that we can convert an Excelfile to a dictionary"""
    tmpdir.chdir()
    defaultvalues = pd.DataFrame()
    # pylint: disable=abstract-class-instantiated
    writer = pd.ExcelWriter("designinput.xlsx", engine="openpyxl")
    MOCK_GENERAL_INPUT.to_excel(
        writer, sheet_name="general_input", index=False, header=None
    )
    MOCK_DESIGNINPUT.to_excel(
        writer, sheet_name="designinput", index=False, header=None
    )
    defaultvalues.to_excel(writer, sheet_name="defaultvalues", index=False, header=None)
    writer.save()

    dict_design = excel2dict_design("designinput.xlsx")
    assert isinstance(dict_design, dict)
    assert dict_design["designtype"] == "onebyone"
    assert dict_design["distribution_seed"] is None
    assert "defaultvalues" in dict_design

    assert isinstance(dict_design["defaultvalues"], dict)
    assert not dict_design["defaultvalues"]  # (it is empty)

    assert isinstance(dict_design["sensitivities"], dict)

    sens = dict_design["sensitivities"]
    # This now contains a key for each sensitivity to make
    assert "rms_seed" in sens
    assert isinstance(sens["rms_seed"], dict)
    assert sens["rms_seed"]["seedname"] == "RMS_SEED"
    assert sens["rms_seed"]["senstype"] == "seed"  # upper-cased

    # Check that we can vary some strings
    writer = pd.ExcelWriter("designinput2.xlsx", engine="openpyxl")
    MOCK_GENERAL_INPUT.to_excel(
        writer, sheet_name="Generalinput", index=False, header=None
    )
    MOCK_DESIGNINPUT.to_excel(
        writer, sheet_name="Design_input", index=False, header=None
    )
    defaultvalues.to_excel(writer, sheet_name="DefaultValues", index=False, header=None)
    writer.save()

    dict_design = excel2dict_design("designinput2.xlsx")
    assert isinstance(dict_design, dict)
    assert dict_design["sensitivities"]["rms_seed"]["senstype"] == "seed"

    # Dump to yaml:
    inputdict_to_yaml(dict_design, "dictdesign.yaml")
    assert os.path.exists("dictdesign.yaml")
    assert "RMS_SEED" in "".join(open("dictdesign.yaml").readlines())


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_duplicate_sensname_exception(tmpdir):
    """Test that exceptions are raised for erroneous sensnames"""
    # pylint: disable=abstract-class-instantiated
    mock_erroneous_designinput = pd.DataFrame(
        data=[
            ["sensname", "numreal", "type", "param_name"],
            ["rms_seed", "", "seed"],
            ["rms_seed", "", "seed"],
        ]
    )
    tmpdir.chdir()
    defaultvalues = pd.DataFrame()

    writer = pd.ExcelWriter("designinput3.xlsx", engine="openpyxl")
    MOCK_GENERAL_INPUT.to_excel(
        writer, sheet_name="general_input", index=False, header=None
    )
    mock_erroneous_designinput.to_excel(
        writer, sheet_name="designinput", index=False, header=None
    )
    defaultvalues.to_excel(writer, sheet_name="defaultvalues", index=False, header=None)
    writer.save()

    with pytest.raises(
        ValueError, match="Two sensitivities can not share the same sensname"
    ):
        excel2dict_design("designinput3.xlsx")


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_strip_spaces(tmpdir):
    """Spaces before and after parameter names are probabaly
    invisible user errors in Excel sheets. Remove them."""
    # pylint: disable=abstract-class-instantiated
    mock_spacious_designinput = pd.DataFrame(
        data=[
            ["sensname", "numreal", "type", "param_name"],
            ["rms_seed   ", "", "seed"],
        ]
    )
    defaultvalues_spacious = pd.DataFrame(
        data=[
            ["parametername", "value"],
            ["  spacious_multiplier", 1.2],
            ["spacious2  ", 3.3],
        ]
    )
    tmpdir.chdir()
    writer = pd.ExcelWriter("designinput_spaces.xlsx", engine="openpyxl")
    MOCK_GENERAL_INPUT.to_excel(
        writer, sheet_name="general_input", index=False, header=None
    )
    mock_spacious_designinput.to_excel(
        writer, sheet_name="designinput", index=False, header=None
    )
    defaultvalues_spacious.to_excel(
        writer, sheet_name="defaultvalues", index=False, header=None
    )
    writer.save()

    dict_design = excel2dict_design("designinput_spaces.xlsx")
    assert list(dict_design["sensitivities"].keys())[0] == "rms_seed"

    # Check default values parameter names:
    def_params = list(dict_design["defaultvalues"].keys())
    assert [par.strip() for par in def_params] == def_params


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_mixed_senstype_exception(tmpdir):
    """Test that exceptions are raised for mixups in user input on types"""
    # pylint: disable=abstract-class-instantiated
    mock_erroneous_designinput = pd.DataFrame(
        data=[
            ["sensname", "numreal", "type", "param_name"],
            ["rms_seed", "", "seed"],
            ["", "", "dist"],
        ]
    )
    tmpdir.chdir()
    defaultvalues = pd.DataFrame()

    writer = pd.ExcelWriter("designinput4.xlsx", engine="openpyxl")
    MOCK_GENERAL_INPUT.to_excel(
        writer, sheet_name="general_input", index=False, header=None
    )
    mock_erroneous_designinput.to_excel(
        writer, sheet_name="designinput", index=False, header=None
    )
    defaultvalues.to_excel(writer, sheet_name="defaultvalues", index=False, header=None)
    writer.save()

    with pytest.raises(ValueError, match="contains more than one sensitivity type"):
        excel2dict_design("designinput4.xlsx")


def test_has_value():
    """Test a function that is used to check if xlsx-cells are empty or not"""
    assert _has_value(1)
    assert not _has_value(np.nan)

    # This possibly makes no sense, but is the current implementation:
    assert _has_value(None)


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_background_sheet(tmpdir):
    """Test loading background values from a sheet"""
    tmpdir.chdir()
    general_input = pd.DataFrame(
        data=[
            ["designtype", "onebyone"],
            ["repeats", 3],
            ["rms_seeds", "default"],
            ["background", "backgroundsheet"],
            ["distribution_seed", "None"],
        ]
    )
    defaultvalues = pd.DataFrame(
        columns=["param_name", "default_value"], data=[["extraseed", "0"]]
    )
    background = pd.DataFrame(
        data=[
            ["param_name", "dist_name", "dist_param1"],
            ["extraseed", "scenario", "30,40,50"],
        ]
    )

    writer = pd.ExcelWriter("designinput.xlsx", engine="openpyxl")
    general_input.to_excel(writer, sheet_name="general_input", index=False, header=None)
    MOCK_DESIGNINPUT.to_excel(
        writer, sheet_name="design_input", index=False, header=None
    )
    defaultvalues.to_excel(writer, sheet_name="defaultvalues", index=False)
    background.to_excel(writer, sheet_name="backgroundsheet", index=False, header=None)
    writer.save()

    dict_design = excel2dict_design("designinput.xlsx")

    # Assert it has been interpreted correctly from input files:
    assert dict_design["background"]["parameters"]["extraseed"] == [
        "scenario",
        ["30,40,50"],
        None,
    ]
    assert dict_design["repeats"] == 3
    assert dict_design["defaultvalues"]["extraseed"] == 0



"""
Used for pre and processing of ert sensitivities, such as
setting up design matrix to run single sensitivities with ERT and
post processing of results to create input for a tornado plot.
Output of this module can be used in custom standalone applications.
"""

from fmu.tools.sensitivities._designsummary import summarize_design
from fmu.tools.sensitivities._tornado_onebyone import calc_tornadoinput
from fmu.tools.sensitivities.create_design import DesignMatrix
from fmu.tools.sensitivities._excel2dict import excel2dict_design, inputdict_to_yaml

__all__ = [
    "summarize_design",
    "calc_tornadoinput",
    "DesignMatrix",
    "excel2dict_design",
    "inputdict_to_yaml",
]



"""Testing qcdata loading of XTGeo data"""

from os.path import abspath

import pytest
import xtgeo
from fmu.tools.qcdata import QCData

# filedata
PATH = abspath(".")  # normally not needed; here due to pytest fixture tmpdir
GRIDFILE = "../xtgeo-testdata/3dgrids/reek/reek_sim_grid.roff"
ZONENAME = "Zone"
ZONEFILE = "../xtgeo-testdata/3dgrids/reek/reek_sim_zone.roff"
WELLFILES = [
    "../xtgeo-testdata/wells/reek/1/OP*.w",
    "../xtgeo-testdata/wells/reek/1/WI*.w",
]

ZONELOGNAME = "Zonelog"
PERFLOGNAME = "Perflog"
REPORT = abspath("./somefile.csv")

DATA1 = {
    "verbosity": "info",
    "path": PATH,
    "grid": GRIDFILE,
    "gridprops": [[ZONENAME, ZONEFILE]],
    "wells": WELLFILES,
}


def test_qcdata():
    """Testing getting data with _QCForwardData class"""

    qcdata = QCData()
    qcdata.parse(data=DATA1)

    assert isinstance(qcdata, QCData)
    assert isinstance(qcdata.grid, xtgeo.Grid)
    assert isinstance(qcdata.gridprops, xtgeo.GridProperties)
    assert isinstance(qcdata.wells, xtgeo.Wells)

    assert qcdata._project is None
    assert qcdata.grid.ncol == 40

    zone = qcdata.gridprops.get_prop_by_name(ZONENAME)
    assert isinstance(zone, xtgeo.GridProperty)
    assert zone.name == ZONENAME
    assert zone.values.mean() == pytest.approx(1.92773, abs=0.01)
    assert zone.ncol == 40

    op1 = qcdata.wells.get_well("OP_1")

    assert ZONELOGNAME in op1.dataframe.columns



# -*- coding: utf-8 -*-

"""Top-level package for fmu-tools"""

from fmu.tools.qcforward.qcforward import wellzonation_vs_grid  # noqa
from fmu.tools.qcproperties.qcproperties import QCProperties  # noqa
from fmu.tools.extract_grid_zone_tops import extract_grid_zone_tops  # noqa

try:
    import roxar  # noqa

    ROXAR = True
except ImportError:
    ROXAR = False

if not ROXAR:

    from .rms import volumetrics  # noqa

    from .sensitivities import DesignMatrix  # noqa
    from .sensitivities import summarize_design  # noqa
    from .sensitivities import calc_tornadoinput  # noqa
    from .sensitivities import excel2dict_design  # noqa

try:
    from .version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"



"""The FipMapper class, mapping region/zones in RMS to FIPxxx in Eclipse."""

from pathlib import Path
import logging
import itertools
import collections
from typing import Union, Dict, List, Any, Optional, Tuple

import yaml
import pandas as pd
from disjoint_set import DisjointSet

logger = logging.getLogger(__name__)


class FipMapper:
    def __init__(
        self,
        *,
        yamlfile: Union[str, Path] = None,
        mapdata: Dict[str, str] = None,
        skipstring: Union[List[str], str] = None,
    ):
        """Represent mappings between region/zones to FIPxxx.

        FipMapper is a class to represent a to map between
        regions/zones in the geomodel (RMS) and to different region partitions
        in the dynamic model (Eclipse).

        Primary usage is to determine which RMS regions corresponds to
        which FIPNUMs, similarly for zones, and in both directions.

        Configuration is via a yaml-file or directly with a dictionary. The
        map can be configured in any direction.

        Several data structures in the dictionary can be used, such that
        the needed information can be extracted from the global configurations
        file.

        Assumptions:
            * A FIPNUM always maps to an arbitrary-length list of regions
            * A FIPNUM always maps to an arbitrary-length list of zones
            * A region always maps to an arbitrary-length list of FIPNUMs
            * A zone always maps to an arbitrary-length list of FIPNUMs
            * A region is assumed to be present for all zones, but not
              relevant in this class
            * A zone is assumed to be present for all regions, but not
              relevant in this class

        For FIPNUM, the datatype must be integers, but can be initialized
        from strings as long as they can be parsed as strings.

        For Region and Zone, a string datatype is assumed, but the yaml
        input is allowed to be integers or integers and strings mixed. Some
        functions will return these as integers if they were inputted as such,
        but at least the disjoint_sets() function will always return
        these as strings.

        Args:
            yamlfile: Filename
            mapdata: direct dictionary input. Provide only one of the
                arguments, not both.
            skipstring: List of strings which will be ignored (e.g. ["Totals"]).
        """
        self._mapdata: Dict[str, dict] = {}  # To be filled with data we need.

        if skipstring is None:
            self.skipstring = []
        if isinstance(skipstring, str):
            self.skipstring = [skipstring]

        if yamlfile is not None and mapdata is not None:
            raise ValueError(
                "Initialize with either yamlfile or explicit data, not both"
            )
        if yamlfile is None and mapdata is None:
            logger.warning("FipMapper initialized with no data")

        if yamlfile is not None:
            logger.info("Loading data from %s", yamlfile)
            with open(yamlfile) as stream:
                yamldata = yaml.safe_load(stream)
            logger.debug(str(yamldata))
        else:
            yamldata = mapdata

        if yamldata is not None:
            self._get_explicit_mapdata(yamldata)

        if yamldata is not None and "global" in yamldata:
            # This is a fmu-config file.
            self._fipdata_from_fmuconfigyaml(yamldata)

        # Webviz YML format:
        if yamldata is not None and "FIPNUM" in yamldata:
            self._fipdata_from_webvizyaml(yamldata)

        assert isinstance(self._mapdata, dict), "FipMapper needs a dictionary"

        # Determine our capabilities:
        self.has_fip2region = "fipnum2region" in self._mapdata
        self.has_fip2zone = "fipnum2zone" in self._mapdata
        self.has_region2fip = "region2fipnum" in self._mapdata
        self.has_zone2fip = "zone2fipnum" in self._mapdata

        # Validate that all FIPNUMs are integers:
        try:
            [int(fip) for fip in self.get_fipnums()]
        except AssertionError:
            # This is for partially empty fipmappers.
            pass
        except ValueError:
            raise TypeError(f"All FIPNUMs must be integers, got {self.get_fipnums}")

    def _get_explicit_mapdata(self, yamldata: Dict[str, Any]):
        """Fetch explicit mapping configuration from a dictionary.

        Set internal flags when maps are found

        Invert maps when possible/needed

        Args:
            yamldata (dict): Configuration object with predefined items
                at the first level.
        """
        if self._mapdata is None:
            self._mapdata = {}
        if "fipnum2region" in yamldata:
            self._mapdata["fipnum2region"] = yamldata["fipnum2region"]
            if "region2fipnum" not in yamldata:
                self._mapdata["region2fipnum"] = invert_map(
                    self._mapdata["fipnum2region"], skipstring=self.skipstring
                )
            self.has_fip2region = True
            self.has_region2fip = True

        if "region2fipnum" in yamldata:
            self._mapdata["region2fipnum"] = yamldata["region2fipnum"]
            if "fipnum2region" not in yamldata:
                logger.debug(self._mapdata["region2fipnum"])
                self._mapdata["fipnum2region"] = invert_map(
                    self._mapdata["region2fipnum"], skipstring=self.skipstring
                )
            self.has_fip2region = True
            self.has_region2fip = True

        if "fipnum2zone" in yamldata:
            self._mapdata["fipnum2zone"] = yamldata["fipnum2zone"]
            if "zone2fipnum" not in yamldata:
                self._mapdata["zone2fipnum"] = invert_map(
                    self._mapdata["fipnum2zone"], skipstring=self.skipstring
                )
            self.has_fip2zone = True
            self.has_zone2fip = True

        if "zone2fipnum" in yamldata:
            self._mapdata["zone2fipnum"] = yamldata["zone2fipnum"]
            if "fip2zone" not in yamldata:
                self._mapdata["fipnum2zone"] = invert_map(
                    self._mapdata["zone2fipnum"], skipstring=self.skipstring
                )
            self.has_fip2zone = True
            self.has_zone2fip = True

    def _fipdata_from_fmuconfigyaml(self, yamldict: dict):
        """This function should be able to build mapping from region/zones to
        FIPNUM based on data it finds in a fmu-config global_master_config.yml
        file.

        How that map should be deduced is not yet defined, and we only support
        having the explicit maps "region2fipnum" etc under the global section

        Args:
            yamldict (dict):
        """
        self._get_explicit_mapdata(yamldict["global"])

    def _fipdata_from_webvizyaml(self, yamldict: dict):
        """This function loads the Webviz yaml syntax for
        REGION/ZONE to FIPNUM mapping,

        The syntax is defined in
        https://github.com/equinor/webviz-subsurface/blob/master/webviz_subsurface/plugins/_reservoir_simulation_timeseries_regional.py#L1422

        Args:
            yamldict (dict):
        """
        self._get_explicit_mapdata(webviz_to_prtvol2csv(yamldict))

    def _fips2regions(self, fips: List[int]) -> List[List[str]]:
        return [self.fip2region(fip_int) for fip_int in fips]

    def get_regions(self) -> List[str]:
        """Obtain a sorted list of the regions that exist in the map"""
        assert "region2fipnum" in self._mapdata, "No data provided for regions"
        try:
            return sorted(self._mapdata["region2fipnum"].keys())
        except TypeError:
            # We get here if some regions are ints and others are strings
            return sorted(map(str, self._mapdata["region2fipnum"].keys()))

    def get_zones(self) -> List[str]:
        """Obtain a sorted list of the zones that exist in the map"""
        assert "zone2fipnum" in self._mapdata, "No data provided for regions"
        try:
            return sorted(self._mapdata["zone2fipnum"].keys())
        except TypeError:
            # We get here if some zones are ints and others are strings
            return sorted(map(str, self._mapdata["zone2fipnum"].keys()))

    def get_fipnums(self) -> List[str]:
        """Obtain a sorted list of the fip numbers that exist in the map"""
        assert "fipnum2region" in self._mapdata, "No data provided for regions"
        return sorted(self._mapdata["fipnum2region"].keys())

    def fip2region(self, fip: int) -> List[str]:
        """Maps one FIP(NUM) integer to list of Region strings. Each FIPNUM
        can map to multiple regions, therefore a list is always returned for
        each FIPNUM.

        Args:
            array: List/array of FIPNUMS, or integer.

        Returns:
            List of strings or list of lists of strings, depending on input.
            Region names that are "integers" will be returned as strings.
            Empty list if no region is known for a specific FIPNUM.
        """
        assert "fipnum2region" in self._mapdata, "No data provided for fip2region"
        try:
            regions = self._mapdata["fipnum2region"][fip]
            if not isinstance(regions, list):
                # Single region in input
                return [regions]
            return regions
        except KeyError:
            logger.warning(
                "Unknown fip %s, known map is %s",
                str(fip),
                str(self._mapdata["fipnum2region"]),
            )
            return []

    def _regions2fips(self, regions: List[str]) -> List[List[int]]:
        return [self.region2fip(region) for region in regions]

    def region2fip(self, region: Union[int, str]) -> List[int]:
        """Maps a Region string/int to FIPNUM(s).

        Args:
            region: Region

        Returns:
            FIPNUM values. None if the region is unknown, many if many FIPNUMs
            are present in the region.
        """
        assert "region2fipnum" in self._mapdata, "No data provided for region2fip"
        if region not in self._mapdata["region2fipnum"]:
            try:
                # If regions have mixed types in yaml, we are sometimes
                # asked for a region as a stringified integer
                region = int(region)
            except ValueError:
                pass
        try:
            fips = self._mapdata["region2fipnum"][region]
            if not isinstance(fips, list):
                # Single FIPNUM in input
                return [int(fips)]
            return [int(fip) for fip in fips]
        except KeyError:
            logger.warning(
                "Unknown region %s, known map is %s",
                str(region),
                str(self._mapdata["region2fipnum"]),
            )
            return []

    def zone2fip(self, zone: Union[str, int]) -> List[int]:
        """Maps a zone to FIPNUMs"""
        assert "zone2fipnum" in self._mapdata, "No data provided for zone2fip"
        if zone not in self._mapdata["zone2fipnum"]:
            try:
                # If zones have mixed types in yaml, we are sometimes
                # asked for a zone as a stringified integer
                zone = int(zone)
            except ValueError:
                pass
        try:
            fips = self._mapdata["zone2fipnum"][zone]
            if not isinstance(fips, list):
                # Single FIPNUM in input
                return [int(fips)]
            return [int(fip) for fip in fips]
        except KeyError:
            logger.warning(
                "Unknown zone %s, known map is %s",
                str(zone),
                str(self._mapdata["zone2fipnum"]),
            )
            return []

    def _fips2zones(self, fips: List[int]) -> List[List[str]]:
        return [self.fip2zone(fip) for fip in fips]

    def fip2zone(self, fip: int) -> List[str]:
        """Maps a FIPNUM integer to an list of Zone strings

        Args:
            array (list): List/array of FIPNUMS, or integer.

        Returns:
            list: Region strings. Always returned as list, and always as
            strings, even if zone "names" are integers. Empty list if no
            zone is assigned to the FIPNUM.
        """
        assert "fipnum2zone" in self._mapdata, "No data provided for fip2zone"
        try:
            zones = self._mapdata["fipnum2zone"][fip]
            if not isinstance(zones, list):
                # Single zone for this FIPNUM
                return [zones]
            return zones
        except KeyError:
            logger.warning("The zone belonging to FIPNUM %s is unknown", str(fip))
            return []  # type: ignore

    def regzone2fip(self, region: str, zone: str) -> List[int]:
        fipreg = self.region2fip(region)
        fipzon = self.zone2fip(zone)
        return sorted(list(set(fipreg).intersection(set(fipzon))))

    def disjoint_sets(self) -> pd.DataFrame:
        """Determine the minimal disjoint sets of a reservoir

        The disjoint sets returned consist of sets that can be split into
        both a set of FIPxxxx list and a region/zone list. Thus, the sum of
        any additive property is comparable on these disjoint sets.

        The returned object is a dataframe that is to be used to group together
        fipnums or regions/zones so they are summable.

        Note that the REGION and ZONE columns always contain strings only, while
        FIPNUM is always an integer.

        Each row represents a cell in the partition where both region, zone and
        fipnum boundaries apply, this the finest possible partition the
        fipmapper data allows. Each row is then assigned to a integer
        identifier in the ``SET`` column. The chosen integers values for each
        set is based on lexiographical sorting of regions, zones and fipnum
        values.

        These sets signifies the minimal grouping of data that must be applied
        in order for volumes in the region/zone partition or fipnum partition
        to be comparable.
        """

        # Generate all possible combinations of the regions and
        # zones we know of:
        regzone_df = pd.DataFrame(
            columns=["REGION", "ZONE"],
            data=itertools.product(self.get_regions(), self.get_zones()),
        )

        # Map all of the region-zone combinations into the accompanying FIPNUMs:
        regzone_df["FIPNUMS"] = regzone_df.apply(
            lambda x: self.regzone2fip(x["REGION"], x["ZONE"]), axis=1
        )

        # The dataframe has lists in the FIPNUMS column when a reg/zone maps
        # to multiple FIPNUMs. Unroll these into one row pr linked FIPNUM:
        dframe = _expand_regzone_df(regzone_df)

        # The `dframe` now has one row pr. smallest "cell" that is interesting
        # in the current context. In some sense, the "intersection" of all
        # possible partitions.

        # Create a dataframe of all possible combinations of these smallest cells:
        edges = pd.merge(
            dframe.assign(dummy=1), dframe.assign(dummy=1), on="dummy"
        ).drop("dummy", axis=1)
        # When Pandas 1.2 is ubiqutous, replace the above statement with:
        # edges = dframe.merge(dframe, how="cross")

        # A partition is equivalent to an equivalence relation.

        # Apply an equivalence relation to the cell combinations:
        edges["NEIGHBOURS"] = edges.apply(
            lambda x: _equivalent_cells(
                x["REGION_x"],
                x["ZONE_x"],
                x["FIPNUM_x"],
                x["REGION_y"],
                x["ZONE_y"],
                x["FIPNUM_y"],
            ),
            axis=1,
        )
        # Filter to only edges that determine which cell linkages
        # that should be grouped:
        neighbourlist = edges[edges["NEIGHBOURS"]]

        # Construct a disjoint set object of all the smallest cells that are
        # to be grouped/unionized:
        ds: DisjointSet = DisjointSet()
        for _, row in dframe.iterrows():
            ds.find((row["REGION"], row["ZONE"], row["FIPNUM"]))

        # Apply the union-find algorithm to determine the partition
        # where all equivalene relations are obeyed:
        for _, pair in neighbourlist.iterrows():
            ds.union(
                (pair["REGION_x"], pair["ZONE_x"], pair["FIPNUM_x"]),
                (pair["REGION_y"], pair["ZONE_y"], pair["FIPNUM_y"]),
            )

        # The union-find algorithm has now "named" each of the components
        # in the disjoint set by a somewhat random mother/root node. This root
        # not is not any more a root compared to the other cells in the set,
        # so each set is instead mapped to consecutive integers.
        id_dict: dict = collections.defaultdict(lambda: len(id_dict))
        dframe["SET"] = [
            id_dict[root]
            for root in dframe.sort_values(["REGION", "ZONE", "FIPNUM"]).apply(
                lambda x: ds.find((x["REGION"], x["ZONE"], x["FIPNUM"])), axis=1
            )
        ]
        dframe["REGION"] = dframe["REGION"].astype(str)
        dframe["ZONE"] = dframe["ZONE"].astype(str)
        dframe["FIPNUM"] = dframe["FIPNUM"].astype(int)
        return dframe


def _equivalent_cells(
    reg1: Any, zon1: Any, fip1: Any, reg2: Any, zon2: Any, fip2: Any
) -> bool:
    """Define the equivalence relation for the reg-zone-fip reservoir partition

    A pair of reg-zone-fip is in the same group if they must be treated together
    when (and have properties summed). Say if you have a value for
    a specific region, but this region contains two FIPNUMs. Then we can never
    treat these two FIPNUMs separately, they must be summed in order to be
    comparable to the value for the region
    """
    return ((reg1 == reg2) and (zon1 == zon2)) or (fip1 == fip2)


def regions_in_set(dframe: pd.DataFrame) -> Dict[int, List[str]]:
    """From the dataframe returned by disjoint_sets(), compute
    a dictionary to map from a set index to a list of regions
    that are members of that set index

    Args:
        dframe: The dataframe emitted by disjoint_sets()
    """
    if dframe.empty:
        return {}
    return (
        dframe.groupby("SET")["REGION"].apply(set).apply(list).apply(sorted).to_dict()
    )


def zones_in_set(dframe: pd.DataFrame) -> Dict[int, List[str]]:
    """From the dataframe returned by disjoint_sets(), compute
    a dictionary to map from a set index to a list of zones
    that are members of that set index

    Args:
        dframe: The dataframe emitted by disjoint_sets()
    """
    if dframe.empty:
        return {}
    return dframe.groupby("SET")["ZONE"].apply(set).apply(list).apply(sorted).to_dict()


def fipnums_in_set(dframe: pd.DataFrame) -> Dict[int, List[int]]:
    """From the dataframe returned by disjoint_sets(), compute
    a dictionary to map from a set index to a list of FIPNUM values
    that are members of that set index

    Args:
        dframe: The dataframe emitted by disjoint_sets()
    """
    if dframe.empty:
        return {}
    return (
        dframe.groupby("SET")["FIPNUM"].apply(set).apply(list).apply(sorted).to_dict()
    )


def regzonefips_in_set(dframe: pd.DataFrame) -> Dict[int, List[Tuple[str, str, int]]]:
    """From the dataframe returned by disjoint_sets(), compute
    a dictionary to map from a set index to a list of tuples
    of the region, zones and fipnums in the set.

    Args:
        dframe: The dataframe emitted by disjoint_sets()
    """
    if dframe.empty:
        return {}
    dframe = dframe.copy()
    dframe["reg-zone-fip"] = dframe[["REGION", "ZONE", "FIPNUM"]].apply(tuple, axis=1)
    return (
        dframe.groupby("SET")["reg-zone-fip"]
        .apply(set)
        .apply(list)
        .apply(sorted)
        .to_dict()
    )


def webviz_to_prtvol2csv(webvizdict: dict):
    """Convert a dict representation of a region/zone map in the Webviz format
    to the prtvol2csv format"""
    if "FIPNUM" in webvizdict and isinstance(webvizdict["FIPNUM"], dict):
        prtvoldict = dict()
        if "groups" in webvizdict["FIPNUM"]:
            if "REGION" in webvizdict["FIPNUM"]["groups"]:
                prtvoldict["region2fipnum"] = webvizdict["FIPNUM"]["groups"]["REGION"]
            if "ZONE" in webvizdict["FIPNUM"]["groups"]:
                prtvoldict["zone2fipnum"] = webvizdict["FIPNUM"]["groups"]["ZONE"]
        else:
            # The "groups" level might go away:
            if "REGION" in webvizdict["FIPNUM"]:
                prtvoldict["region2fipnum"] = webvizdict["FIPNUM"]["REGION"]
            if "ZONE" in webvizdict["FIPNUM"]:
                prtvoldict["zone2fipnum"] = webvizdict["FIPNUM"]["ZONE"]
        return prtvoldict
    return {}


def invert_map(
    dictmap: Dict[str, Any], skipstring: Optional[Union[list, str]] = None
) -> Dict[str, List[Any]]:
    """Invert a dictionary, supporting many-to-many maps.

    Args:
        dictmap
        skipstring: List of strings which will be ignored (e.g. "Totals").
    """
    if skipstring is None:
        skipstring = []
    if isinstance(skipstring, str):
        skipstring = [skipstring]

    inv_map: Dict[str, List[Any]] = {}
    for key, value in dictmap.items():
        if key in skipstring or value in skipstring:
            continue
        if isinstance(value, list):
            for _value in value:
                inv_map[_value] = list(
                    set(inv_map.get(_value, set())).union(set([key]))
                )
        else:
            base = set(inv_map.get(value, set()))
            # mypy workaround: https://github.com/python/mypy/issues/2013
            inv_map[value] = list(base.union(set([key])))

    for key, value in inv_map.items():
        try:
            inv_map[key] = sorted(list(inv_map[key]))
        except TypeError:
            # Datatype of keys are mixed, typically int and str.
            inv_map[key] = sorted(map(str, list(inv_map[key])))

    return inv_map


def _expand_regzone_df(dframe: pd.DataFrame, fipname: str = "FIPNUM") -> pd.DataFrame:
    """Unroll dataframe rows with a FIPNUM list in the "FIPNUMS" column"""
    new_rows = []
    for _, row in dframe.iterrows():
        for fipnumber in row[fipname + "S"]:
            new_rows.append(
                {
                    "REGION": row["REGION"],
                    "ZONE": row["ZONE"],
                    fipname: fipnumber,
                    "REGZONE": str(row["REGION"]) + "-" + str(row["ZONE"]),
                }
            )
    return pd.DataFrame(new_rows)
